""" Defines store functions for working with Projects.
"""
import arcgis
from arcgis.gis import Item
from arcgis.apps import workforce
from arcgis.apps.workforce._store._definitions import *
from arcgis.apps.workforce.exceptions import WorkforceError
from warnings import warn
import concurrent.futures
import os


def get_project(project_id, gis):
    """Loads and returns a workforce project.
    :param gis: An authenticated arcigs.gis.GIS object.
    :param project_id: The project's id. Version 1 - id of the project item. Version 2 - id of the feature service
    :returns: workforce.Project
    """
    item = Item(gis, project_id)
    return workforce.Project(item)


def create_project(title, summary=None, major_version=None, gis=None):
    """
    Creates a new Workforce Project

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    title                  :class:`String`.
                           The title of the Project to create (must be unique to the organization)
    ------------------     --------------------------------------------------------------------
    summary                :class:`String`.
                           The summary of the Project
    ------------------     --------------------------------------------------------------------
    major_version          Optional :class:`Int`
                           The version of the Project to create. 1 represents the original
                           Workforce Project which does not support offline. 2 represents the newer
                           Workforce Project which supports offline among other things. Defaults
                           to 2 in GIS 8.2 and higher
    ------------------     --------------------------------------------------------------------
    gis                    Optional :class:`~arcgis.gis.GIS`.
                           The authenticated GIS to use.
                           Defaults to the active GIS if None is provided.
    ==================     ====================================================================

    Returns a :class:`arcgis.apps.workforce.Project`
    """

    if gis is None:
        gis = arcgis.env.active_gis

    if major_version is None:
        if gis.version < [8, 2]:
            major_version = 1
        else:
            major_version = 2

    if major_version == 1:
        return _v1_create_project(gis, summary, title)

    elif major_version == 2:
        if gis.version < [7, 1]:
            raise ValueError(
                "Offline Workforce Projects not supported at GIS lower than 7.1"
            )
        return _v2_create_project(gis, summary, title)

    else:
        raise WorkforceError("Invalid Project Version Specified")


def _get_basemap(gis):
    """
    Helper method to get the basemap to use. Tries to use the org defined basemap.
    :param gis:
    :return:
    """
    default_raster_basemap = {
        "baseMapLayers": [
            {
                "id": "World_Topo_Map",
                "layerType": "ArcGISTiledMapServiceLayer",
                "url": "https://services.arcgisonline.com/ArcGIS/rest/services/World_Topo_Map/MapServer",
                "visibility": True,
                "opacity": 1,
                "title": "World_Topo_Map",
            }
        ],
        "title": "Topographic",
    }
    try:
        if gis.properties["useVectorBasemaps"]:
            bm_group = gis.groups.search(
                gis.properties["vectorBasemapGalleryGroupQuery"], outside_org=True
            )[0]
            query = 'group:{} AND name: "0220_Navigation_Title"'.format(bm_group.id)
            items = gis.content.search(query, outside_org=True)
            if items:
                return items[0].get_data()["baseMap"]
        else:
            return default_raster_basemap
    except:
        return default_raster_basemap


def _create_application_properties():
    """
    Helper method to build the application properties
    :return:
    """
    return {
        "viewing": {
            "search": {
                "enabled": True,
                "disablePlaceFinder": True,
                "layers": [
                    {
                        "id": "Workers_0",
                        "field": {
                            "name": "name",
                            "exactMatch": True,
                            "type": "esriFieldTypeString",
                        },
                    }
                ],
            }
        }
    }


def _build_operational_layers(
    item, popup_def=None, visibility=True, layer_index=0, capabilities=None
):
    """
    Helper method to build operational layers
    :param item: The item containing the layer (at index 0)
    :param popup_def: The popup definition of the layer
    :param visibility: Is this layer visible in the map
    :return: The operational layer dictionary
    """
    layer = item.layers[layer_index]
    op_layer = {
        "layerType": "ArcGISFeatureLayer",
        "opacity": 1,
        "itemId": item.id,
        "id": "{}_0".format(layer.properties["name"]),
        "title": layer.properties["name"],
        "url": layer.url,
        "visibility": visibility,
    }
    if capabilities:
        op_layer["capabilities"] = capabilities
    if popup_def:
        op_layer["popupInfo"] = popup_def
    return op_layer


def _build_table(item, table_index):
    """
    Helper method to build table
    :param item: The item containing the layer (at index 0)
    :param table_index: The index for the table in the item
    """
    table = item.tables[table_index]
    op_table = {
        "capabilities": "Create,Delete,Query,Update,Editing,Sync",
        "url": table.url,
        "id": "{}_0".format(table.properties["name"]),
        "title": table.properties["name"],
        "itemId": item.id,
    }
    return op_table


def _v2_create_project(gis, summary, title):
    """Creates project following version 2 Workforce schema"""
    for f in gis.users.me.folders:
        if f["title"].lower() == title.lower():
            raise WorkforceError("A folder named '{}' already exists.".format(title))

    with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
        future_folder = executor.submit(gis.content.create_folder, title)
        future_group = executor.submit(
            gis.groups.create,
            title,
            "workforce",
            access="private",
            is_invitation_only=True,
            is_view_only=True,
            sort_field="title",
            sort_order="asc",
        )
    future_folder.result()
    group = future_group.result()
    group.protected = True
    folder_name = title
    group_id = group.id
    workforce_service_name = "workforce_{}".format(group_id)
    workforce_service_item = _v2_create_service_with_layers(
        gis,
        folder_name,
        workforce_service_name,
        assignment_layer_definition_v2,
        worker_layer_definition_v2,
        dispatcher_table_definition_v2,
        assignment_type_table_definition_v2,
        app_integration_table_definition_v2,
        title,
    )

    # create webmaps
    my_path = os.path.abspath(os.path.dirname(__file__))
    thumbnail = os.path.join(
        my_path, "/".join(("resources", "default-project-thumbnail.png"))
    )
    with concurrent.futures.ThreadPoolExecutor(max_workers=2) as executor:
        workers_webmap_future = executor.submit(
            _v2_create_worker_webmap,
            gis,
            folder_name,
            workforce_service_item,
            assignment_layer_popup_definition_v2,
            worker_layer_popup_definition_v2,
            title,
            summary,
            thumbnail,
        )
        dispatchers_webmap_future = executor.submit(
            _v2_create_dispatcher_webmap,
            gis,
            folder_name,
            workforce_service_item,
            assignment_layer_popup_definition_v2,
            worker_layer_popup_definition_v2,
            title,
            summary,
            thumbnail,
        )
    workers_webmap = workers_webmap_future.result()
    dispatchers_webmap = dispatchers_webmap_future.result()

    project_items = [workforce_service_item, workers_webmap, dispatchers_webmap]
    # share and protect items
    for i in project_items:
        i.share(groups=[group])
        i.protect()

    # set fs item properties / thumbnail
    workforce_service_item.update(
        thumbnail=thumbnail,
        item_properties={
            "snippet": summary,
            "properties": {
                "workforceProjectGroupId": group_id,
                "workforceProjectVersion": "2.0.0",
                "workforceDispatcherMapId": dispatchers_webmap.id,
                "workforceWorkerMapId": workers_webmap.id,
            },
        },
    )

    # manually add the owner as the first dispatcher
    user_id_field_name = "userId"
    for field in workforce_service_item.tables[0].properties.fields:
        if field["name"].lower() == "userid":
            user_id_field_name = field["name"]
            break

    workforce_service_item.tables[0].edit_features(
        adds=[
            arcgis.features.Feature(
                attributes={
                    "name": gis.users.me.fullName,
                    user_id_field_name: gis.users.me.username,
                }
            )
        ]
    )

    # add relationships
    try:
        workers_webmap.add_relationship(
            workforce_service_item, "WorkforceMap2FeatureService"
        )
        dispatchers_webmap.add_relationship(
            workforce_service_item, "WorkforceMap2FeatureService"
        )
    except Exception:
        warn(
            "Relationship not added. This version of ArcGIS may not support the WorkforceMap2FeatureService relationship"
        )

    # create the Project to return, add navigator as default integration
    project = arcgis.apps.workforce.Project(workforce_service_item)
    project.integrations.add(
        integration_id="arcgis-navigator",
        prompt="Navigate to Assignment",
        url_template="https://navigator.arcgis.app?stop=${assignment.latitude},${assignment.longitude}&stopname=${assignment.location}&callback=https://workforce.arcgis.app&callbackprompt=Workforce",
    )

    return project


def _v1_create_project_item(
    gis,
    folder_name,
    assignments_item,
    dispatchers_item,
    workers_item,
    tracks_item,
    dispatchers_webmap,
    workers_webmap,
    title,
    summary,
    folder_id,
    group_id,
):
    """
    Creates the project item
    :param gis: An authenticated GIS
    :param folder_name: The name of the folder in which to place the project
    :param assignments_item: The assignments item
    :param dispatchers_item: THe dispatchers item
    :param workers_item: The workers item
    :param tracks_item: The tracks item
    :param dispatchers_webmap: The dispatcher webmap item
    :param workers_webmap: The workers webmap item
    :param title: The title of the project
    :param summary: The summary of the project
    :param folder_id: The folder id
    :param group_id: The group id
    :return: The project item
    """
    item_properties = {
        "title": title,
        "snippet": summary,
        "tags": "workforce",
        "type": "Workforce Project",
        "typeKeywords": "Workforce Project",
    }
    project_data = {
        "workerWebMapId": workers_webmap.id,
        "dispatcherWebMapId": dispatchers_webmap.id,
        "dispatchers": {
            "serviceItemId": dispatchers_item.id,
            "url": dispatchers_item.layers[0].url,
        },
        "assignments": {
            "serviceItemId": assignments_item.id,
            "url": assignments_item.layers[0].url,
        },
        "workers": {
            "serviceItemId": workers_item.id,
            "url": workers_item.layers[0].url,
        },
        "tracks": {
            "serviceItemId": tracks_item.id,
            "url": tracks_item.layers[0].url,
            "enabled": False,
            "updateInterval": 30,
        },
        "assignmentIntegrations": [
            {
                "id": "default-navigator",
                "prompt": "Navigate to Assignment",
                "urlTemplate": "https://navigator.arcgis.app?stop=${assignment.latitude},${assignment.longitude}&stopname=${assignment.location}&callback=https://workforce.arcgis.app&callbackprompt=Workforce",
            }
        ],
        "version": "1.3.0",
        "groupId": group_id,
        "folderId": folder_id,
    }
    item_properties["text"] = json.dumps(project_data)
    item = gis.content.add(item_properties, folder=folder_name)
    return item


def _v2_create_worker_webmap(
    gis,
    folder_name,
    workforce_service_item,
    assignments_popup_def,
    workers_popup_def,
    title,
    summary,
    thumbnail,
):
    """
    Creates the worker webmap
    :param gis: An authenticated GIS
    :param folder_name: The name of the folder in which to place the webmap
    :param workforce_service_item: The workforce service item
    :param assignments_popup_def: The assignments popup definition
    :param workers_popup_def: The workers popup definition
    :param title: The title of the project
    :return: The worker webmap item
    """
    extent = _get_default_extent(gis)
    array_extent = [[extent["xmin"], extent["ymin"]], [extent["xmax"], extent["ymax"]]]
    item_properties = {
        "title": "{}".format(title),
        "snippet": summary,
        "extent": array_extent,
        "type": "Web Map",
        "typeKeywords": "ArcGIS Online,Explorer Web Map,Map,Offline,Online Map,Web Map,Workforce Worker,Data Editing",
        "properties": {"workforceFeatureServiceId": workforce_service_item.id},
    }

    webmap_data = {
        "baseMap": _get_basemap(gis),
        "operationalLayers": [],
        "spatialReference": {"wkid": 102100, "latestWkid": 3857},
        "authoringApp": "ArcGISPythonAPI",
        "authoringAppVersion": str(arcgis.__version__),
        "tables": [],
        "applicationProperties": {
            "viewing": {
                "search": {
                    "enabled": True,
                    "disablePlaceFinder": False,
                    "layers": [
                        {
                            "id": "Workers_0",
                            "field": {
                                "name": "name",
                                "exactMatch": False,
                                "type": "esriFieldTypeString",
                            },
                        }
                    ],
                }
            }
        },
        "version": "2.11",
    }
    webmap_data["operationalLayers"].append(
        _build_operational_layers(
            workforce_service_item, assignments_popup_def, layer_index=0
        )
    )
    webmap_data["operationalLayers"].append(
        _build_operational_layers(
            workforce_service_item, workers_popup_def, layer_index=1
        )
    )
    webmap_data["tables"].append(_build_table(workforce_service_item, table_index=0))
    webmap_data["tables"].append(_build_table(workforce_service_item, table_index=1))
    webmap_data["tables"].append(_build_table(workforce_service_item, table_index=2))
    item_properties["text"] = json.dumps(webmap_data)

    item = gis.content.add(item_properties, folder=folder_name, thumbnail=thumbnail)
    return item


def _v2_create_dispatcher_webmap(
    gis,
    folder_name,
    workforce_service_item,
    assignments_popup_def,
    workers_popup_def,
    title,
    summary,
    thumbnail,
):
    """
    Creates the dispatcher webmap
    :param gis: An authenticated GIS
    :param folder_name: The folder in which to store the webmap
    :param workforce_service_item: The workforce service item item
    :param assignments_popup_def: The assignments popup definition
    :param workers_popup_def: The workers popup definition
    :param title: The title of the project
    :return: The dispatcher webmap item
    """
    extent = _get_default_extent(gis)
    array_extent = [[extent["xmin"], extent["ymin"]], [extent["xmax"], extent["ymax"]]]
    item_properties = {
        "title": "{} Dispatcher Map".format(title),
        "snippet": summary,
        "extent": array_extent,
        "type": "Web Map",
        "typeKeywords": "ArcGIS Online,Explorer Web Map,Map,Offline,Online Map,Web Map,Workforce Dispatcher",
        "properties": {"workforceFeatureServiceId": workforce_service_item.id},
    }

    webmap_data = {
        "baseMap": _get_basemap(gis),
        "operationalLayers": [],
        "spatialReference": {"wkid": 102100, "latestWkid": 3857},
        "authoringApp": "ArcGISPythonAPI",
        "authoringAppVersion": str(arcgis.__version__),
        "version": "2.11",
    }
    webmap_data["operationalLayers"].append(
        _build_operational_layers(
            workforce_service_item,
            assignments_popup_def,
            layer_index=0,
            capabilities="Query,Sync",
        )
    )
    webmap_data["operationalLayers"].append(
        _build_operational_layers(
            workforce_service_item,
            workers_popup_def,
            layer_index=1,
            capabilities="Query,Sync",
        )
    )
    item_properties["text"] = json.dumps(webmap_data)
    item = gis.content.add(item_properties, folder=folder_name, thumbnail=thumbnail)
    return item


def _v2_create_service(
    gis, service_name, folder_name, spatial_ref, item_properties=None
):
    """
    Helper method that creates a workforce service
    :param gis: An authenticated GIS
    :param service_name: The name of the service
    :param folder_name: The name of the folder in which to place the service
    :param spatial_ref: The spatial reference to use
    :return: The service item
    """
    capabilities = ["Query", "Editing", "Create", "Update", "Delete", "Sync"]
    if not gis.properties.isPortal:
        capabilities.append("ChangeTracking")
    create_params = {
        "name": service_name,
        "capabilities": ",".join(capabilities),
        "spatialReference": spatial_ref,
        "syncRowsMovedOutsideFilter": True,
        "preserveLayerIds": True,
    }

    return gis.content.create_service(
        service_name,
        folder=folder_name,
        create_params=create_params,
        item_properties=item_properties,
        service_type="featureService",
    )


def _v2_create_service_with_layers(
    gis,
    folder_name,
    service_name,
    assignments_layer_def,
    workers_layer_def,
    dispatchers_table_def,
    assignment_type_table_def,
    integration_table_def,
    title,
):
    """
    Creates a service, adds, and layer and optionally enables attachments
    :param gis: An authenticated GIS
    :param folder_name: The name of the folder in which to place the service
    :param service_name: The name of the service
    :param assignments_layer_def: The assignments layer definition (dictionary)
    :param workers_layer_def: The workers layer definition (dictionary)
    :param dispatchers_table_def: The dispatchers table definition (dictionary)
    :param assignment_type_table_def: The assignment type table definition (dictionary)
    :param integration_table_def: The integration table definition (dictionary)
    :return: The service item
    """
    default_extent = gis.properties["defaultExtent"]
    spatial_reference = default_extent["spatialReference"]
    layer_defs = [assignments_layer_def, workers_layer_def]
    table_defs = [
        dispatchers_table_def,
        assignment_type_table_def,
        integration_table_def,
    ]
    if gis.content.is_service_name_available(service_name, "featureService"):
        item_properties = {
            "title": title,
            "tags": "workforce",
            "typeKeywords": "Workforce Project",
        }
        item = _v2_create_service(
            gis, service_name, folder_name, spatial_reference, item_properties
        )
        feature_layer_collection = arcgis.features.FeatureLayerCollection.fromitem(item)
    else:
        raise Exception("Service name already exists.")

    for layer_def in layer_defs:
        layer_def["extent"] = default_extent

    feature_layer_collection.manager.add_to_definition(
        {"layers": layer_defs, "tables": table_defs}
    )

    # enabled editor tracking
    feature_layer_collection.manager.update_definition(
        {
            "editorTrackingInfo": {
                "enableEditorTracking": True,
                "enableOwnershipAccessControl": False,
                "allowOthersToUpdate": True,
                "allowOthersToDelete": True,
            }
        }
    )
    return item


def _get_default_extent(gis):
    """
    Helper method to get the extent used by the org or use default extent
    :param gis: An authenticated GIS
    :return: A dictionary representing an extent
    """
    default_extent = gis.properties["defaultExtent"]
    try:
        extent = arcgis.geometry.project(
            [arcgis.geometry.Envelope(default_extent)],
            default_extent["spatialReference"]["wkid"],
            4326,
        )[0]
    except:
        extent = {
            "spatialReference": {"wkid": 4326},
            "xmin": -180,
            "ymin": -90,
            "xmax": 180,
            "ymax": 90,
        }
    return extent


def _v1_create_project(gis, summary, title):
    for f in gis.users.me.folders:
        if f["title"].lower() == title.lower():
            raise WorkforceError("A folder named '{}' already exists.".format(title))

    with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
        future_folder = executor.submit(gis.content.create_folder, title)
        future_group = executor.submit(
            gis.groups.create,
            title,
            "workforce",
            access="private",
            is_invitation_only=True,
            is_view_only=True,
            sort_field="title",
            sort_order="asc",
        )
    folder = future_folder.result()
    group = future_group.result()
    folder_name = title
    folder_id = folder["id"]
    group_id = group.id
    # create services
    worker_service_name = "workers_{}".format(group_id)
    dispatcher_service_name = "dispatchers_{}".format(group_id)
    assignment_service_name = "assignments_{}".format(group_id)
    tracking_service_name = "location_{}".format(group_id)
    with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
        workers_item_future = executor.submit(
            _v1_create_service_with_layer,
            gis,
            folder_name,
            worker_service_name,
            worker_layer_definition_v1,
        )
        dispatchers_item_future = executor.submit(
            _v1_create_service_with_layer,
            gis,
            folder_name,
            dispatcher_service_name,
            dispatcher_layer_definition_v1,
        )
        assignments_item_future = executor.submit(
            _v1_create_service_with_layer,
            gis,
            folder_name,
            assignment_service_name,
            assignment_layer_definition_v1,
            attachments=True,
        )
        tracks_item_future = executor.submit(
            _v1_create_service_with_location_tracking_layer,
            gis,
            folder_name,
            tracking_service_name,
            tracking_layer_definition_v1,
        )
    workers_item = workers_item_future.result()
    dispatchers_item = dispatchers_item_future.result()
    assignments_item = assignments_item_future.result()
    tracks_item = tracks_item_future.result()

    # create webmaps
    with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
        workers_webmap_future = executor.submit(
            _v1_create_worker_webmap,
            gis,
            folder_name,
            assignments_item,
            workers_item,
            tracks_item,
            assignment_layer_popup_definition_v1,
            worker_layer_popup_definition_v1,
            title,
        )
        dispatchers_webmap_future = executor.submit(
            _v1_create_dispatcher_webmap,
            gis,
            folder_name,
            assignments_item,
            workers_item,
            assignment_layer_popup_definition_v1,
            worker_layer_popup_definition_v1,
            title,
        )
    workers_webmap = workers_webmap_future.result()
    dispatchers_webmap = dispatchers_webmap_future.result()

    # create the project
    project_item = _v1_create_project_item(
        gis,
        folder_name,
        assignments_item,
        dispatchers_item,
        workers_item,
        tracks_item,
        dispatchers_webmap,
        workers_webmap,
        title,
        summary,
        folder_id,
        group_id,
    )
    project_items = [
        workers_item,
        dispatchers_item,
        assignments_item,
        tracks_item,
        workers_webmap,
        dispatchers_webmap,
        project_item,
    ]
    # share and protect items
    for i in project_items:
        i.share(groups=[group])
        i.protect()

    my_path = os.path.abspath(os.path.dirname(__file__))
    thumbnail = os.path.join(
        my_path, "/".join(("resources", "default-project-thumbnail.png"))
    )
    project_item.update(thumbnail=thumbnail)

    # manually add the owner as the first dispatcher
    user_id_field_name = "userId"
    for field in dispatchers_item.layers[0].properties.fields:
        if field["name"].lower() == "userid":
            user_id_field_name = field["name"]
            break

    dispatchers_item.layers[0].edit_features(
        adds=[
            arcgis.features.Feature(
                attributes={
                    "name": gis.users.me.fullName,
                    user_id_field_name: gis.users.me.username,
                }
            )
        ]
    )

    # create the Project to return
    project = arcgis.apps.workforce.Project(project_item)
    return project


def _v1_create_worker_webmap(
    gis,
    folder_name,
    assignments_item,
    workers_item,
    tracking_item,
    assignments_popup_def,
    workers_popup_def,
    title,
):
    """
    Creates the worker webmap
    :param gis: An authenticated GIS
    :param folder_name: The name of the folder in which to place the webmap
    :param assignments_item: The assignments item
    :param workers_item: The workers item
    :param tracking_item: The tracking item
    :param assignments_popup_def: The assignments popup definition
    :param workers_popup_def: The workers popup definition
    :param title: The title of the project
    :return: The worker webmap item
    """
    extent = _get_default_extent(gis)
    array_extent = [[extent["xmin"], extent["ymin"]], [extent["xmax"], extent["ymax"]]]
    item_properties = {
        "title": "{}_workers".format(title),
        "tags": "workforce-worker",
        "extent": array_extent,
        "type": "Web Map",
        "typeKeywords": "ArcGIS Online,Explorer Web Map,Map,Online Map,Web Map,Workforce Project",
    }

    webmap_data = {
        "baseMap": _get_basemap(gis),
        "operationalLayers": [],
        "spatialReference": {"wkid": 102100, "latestWkid": 3857},
        "authoringApp": "ArcGISPythonAPI",
        "authoringAppVersion": str(arcgis.__version__),
        "applicationProperties": {
            "viewing": {
                "search": {
                    "enabled": True,
                    "disablePlaceFinder": False,
                    "layers": [
                        {
                            "id": "Workers_0",
                            "field": {
                                "name": "name",
                                "exactMatch": False,
                                "type": "esriFieldTypeString",
                            },
                        }
                    ],
                }
            }
        },
        "version": "2.11",
    }

    webmap_data["operationalLayers"].append(
        _build_operational_layers(tracking_item, visibility=False, capabilities="Query")
    )
    webmap_data["operationalLayers"].append(
        _build_operational_layers(
            assignments_item, assignments_popup_def, capabilities="Query,Sync"
        )
    )
    webmap_data["operationalLayers"].append(
        _build_operational_layers(
            workers_item, workers_popup_def, capabilities="Query,Sync"
        )
    )
    item_properties["text"] = json.dumps(webmap_data)

    item = gis.content.add(item_properties, folder=folder_name)
    return item


def _v1_create_dispatcher_webmap(
    gis,
    folder_name,
    assignments_item,
    workers_item,
    assignments_popup_def,
    workers_popup_def,
    title,
):
    """
    Creates the dispatcher webmap
    :param gis: An authenticated GIS
    :param folder_name: The folder in which to store the webmap
    :param assignments_item: The assignments item
    :param workers_item: The workers item
    :param assignments_popup_def: The assignments popup definition
    :param workers_popup_def: The workers popup definition
    :param title: The title of the project
    :return: The dispatcher webmap item
    """
    extent = _get_default_extent(gis)
    array_extent = [[extent["xmin"], extent["ymin"]], [extent["xmax"], extent["ymax"]]]
    item_properties = {
        "title": "{}_dispatchers".format(title),
        "tags": "workforce-dispatcher",
        "extent": array_extent,
        "type": "Web Map",
        "typeKeywords": "ArcGIS Online,Explorer Web Map,Map,Offline,Online Map,Web Map,Workforce Project",
    }

    webmap_data = {
        "baseMap": _get_basemap(gis),
        "operationalLayers": [],
        "spatialReference": {"wkid": 102100, "latestWkid": 3857},
        "authoringApp": "ArcGISPythonAPI",
        "authoringAppVersion": str(arcgis.__version__),
        "version": "2.11",
    }
    webmap_data["operationalLayers"].append(
        _build_operational_layers(
            assignments_item, assignments_popup_def, capabilities="Query,Sync"
        )
    )
    webmap_data["operationalLayers"].append(
        _build_operational_layers(
            workers_item, workers_popup_def, capabilities="Query,Sync"
        )
    )
    item_properties["text"] = json.dumps(webmap_data)
    item = gis.content.add(item_properties, folder=folder_name)
    return item


def _v1_create_service_with_layer(
    gis, folder_name, service_name, layer_definition, attachments=False, sync=True
):
    """
    Creates a service, adds, and layer and optionally enables attachments
    :param gis: An authenticated GIS
    :param folder_name: The name of the folder in which to place the service
    :param service_name: The name of the service
    :param layer_definition: The definition of the layer (dictionary)
    :param attachments: Should attachments be enabled on the service
    :return: The service item
    """
    default_extent = gis.properties["defaultExtent"]
    spatial_reference = default_extent["spatialReference"]
    layer_definition["extent"] = default_extent
    if gis.content.is_service_name_available(service_name, "featureService"):
        item = _v1_create_service(
            gis, service_name, folder_name, spatial_reference, sync=sync
        )
        item.update({"tags": "workforce"})
        feature_layer_collection = arcgis.features.FeatureLayerCollection.fromitem(item)

        # add layer
        feature_layer_collection.manager.add_to_definition(
            {"layers": [layer_definition]}
        )

        # enabled editor tracking
        feature_layer_collection.manager.update_definition(
            {
                "editorTrackingInfo": {
                    "enableEditorTracking": True,
                    "enableOwnershipAccessControl": False,
                    "allowOthersToUpdate": True,
                    "allowOthersToDelete": True,
                }
            }
        )

        # enable attachments
        if attachments:
            feature_layer_collection.layers[0].manager.update_definition(
                {"hasAttachments": True}
            )
        return item
    else:
        raise Exception("Service name already exists.")


def _v1_create_service_with_location_tracking_layer(
    gis, folder_name, service_name, layer_definition
):
    """
    Creates the location tracking service, adds a layer and optionally enables attachments
    :param gis: An authenticated GIS
    :param folder_name: The name of the folder in which to place the service
    :param service_name: The name of the service
    :param layer_definition: The definition of the layer (dictionary)
    :return: The location tracking item
    """
    item = _v1_create_service_with_layer(
        gis, folder_name, service_name, layer_definition, sync=False
    )
    item.layers[0].manager.update_definition(
        {
            "timeInfo": {
                "startTimeField": item.layers[0].properties["editFieldsInfo"][
                    "creationDateField"
                ],
                "timeReference": {"timeZone": "UTC", "respectDaylightSaving": False},
                "timeInterval": 0,
                "exportOptions": {
                    "useTime": True,
                    "timeDataCumulative": True,
                    "TimeOffset": 0,
                    "timeOffsetUnits": "esriTimeUnitsCenturies",
                },
                "hasLiveData": True,
            }
        }
    )
    item.update(
        {
            "tags": "workforce, Location Tracking",
            "typeKeywords": "Collector,Data,Feature Service Template,Layer,Layer Template,Location Tracking,Platform,Service template,Template",
        }
    )
    return item


def _v1_create_service(gis, service_name, folder_name, spatial_ref, sync=True):
    """
    Helper method that creates a workforce service
    :param gis: An authenticated GIS
    :param service_name: The name of the service
    :param folder_name: The name of the folder in which to place the service
    :param spatial_ref: The spatial reference to use
    :return: The service item
    """
    create_params = {
        "name": service_name,
        "allowGeometryUpdates": True,
        "units": "esriMeters",
        "spatialReference": spatial_ref,
        "capabilities": "Query,Editing,Create,Update,Delete",
    }
    if sync:
        create_params["syncEnabled"] = True
        create_params["syncCapabilities"] = {
            "supportsAsync": True,
            "supportsRegisteringExistingData": True,
            "supportsSyncDirectionControl": True,
            "supportsPerLayerSync": True,
            "supportsPerReplicaSync": True,
            "supportsSyncModelNone": True,
            "supportsRollbackOnFailure": True,
        }
        create_params["capabilities"] = "Query,Editing,Create,Update,Delete,Sync"

    return gis.content.create_service(
        service_name,
        folder=folder_name,
        create_params=create_params,
        service_type="featureService",
    )



import arcgis
import datetime as _dt
from arcgis._impl.common._utils import _lazy_property
from arcgis.apps.tracker import LocationTrackingError


class LocationTrackingManager:
    """
    This manages Location Sharing for an organization.
    It can be accessed from the gis as `location_tracking`

    Additional information can be found `here <https://doc.arcgis.com/en/tracker/help/configure-location-tracking.htm>`_

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    gis                    Required :class:`~arcgis.gis.GIS`. The GIS to configure location
                           sharing for.
    ==================     ====================================================================
    """

    def __init__(self, gis):
        self._gis = gis

    def _validate_environment(self, location_tracking_enabled=True):
        if self._gis is None:
            raise LocationTrackingError("You must use a valid GIS")
        if float(self._gis.properties.get("currentVersion", "0")) < 7.1:
            raise LocationTrackingError(
                "Location Tracking requires ArcGIS Enterprise 10.7+ or ArcGIS Online"
            )
        if self._gis.users.me.role != "org_admin":
            raise LocationTrackingError(
                "You must be an administrator to configure Location Tracking"
            )
        if self._gis.properties.get("isPortal", False) and not self._gis.properties.get(
            "hasSpatioTemporalArcGISDataStore", False
        ):
            raise LocationTrackingError(
                "Location Tracking requires the Spatiotemporal Big Data Store"
            )
        if (
            location_tracking_enabled
            and "locationTracking" not in self._gis.properties.helperServices
        ):
            raise LocationTrackingError("Location Tracking is not enabled.")
        if (
            not location_tracking_enabled
            and "locationTracking" in self._gis.properties.helperServices
        ):
            raise LocationTrackingError("Location Tracking is already enabled.")

    @property
    def _use_location_sharing(self):
        return float(self._gis.properties.get("currentVersion", "0")) >= 10.1

    def enable(
        self,
        tracks_layer_shards: int = 6,
        lkl_layer_shards: int = 3,
        tracks_layer_rolling_index_strategy: str = "Monthly",
    ):
        """
        Enables location sharing for the organization.

        ===================================       ===============================================================
        **Parameter**                              **Description**
        -----------------------------------       ---------------------------------------------------------------
        tracks_layer_shards                       The number of shards to use for the tracks layer. This only
                                                  applies for ArcGIS Enterprise.
        -----------------------------------       ---------------------------------------------------------------
        lkl_layer_shards                          The number of shards to use for the last known location layer.
                                                  This only applies for ArcGIS Enterprise.
        -----------------------------------       ---------------------------------------------------------------
        tracks_layer_rolling_index_strategy       The rolling index strategy for the tracks layer
                                                  ["Daily", "Weekly", "Monthly", "Yearly", "Decade", "Century"]
                                                  This only applies for ArcGIS Enterprise.
        ===================================       ===============================================================

        :return: True if successful, False otherwise
        """
        if self.status == "enabled":
            return False
        self._validate_environment(location_tracking_enabled=False)
        if tracks_layer_rolling_index_strategy not in [
            "Daily",
            "Weekly",
            "Monthly",
            "Yearly",
            "Decade",
            "Century",
        ]:
            raise ValueError(
                f"Invalid rolling index strategy '{tracks_layer_rolling_index_strategy}'"
            )
        if float(
            self._gis.properties.get("currentVersion", "0")
        ) <= 7.3 and tracks_layer_rolling_index_strategy in ["Century", "Decade"]:
            raise ValueError(
                f"'{tracks_layer_rolling_index_strategy}' is not supported for this version of Enterprise"
            )
        folder_title = (
            "Location Sharing" if self._use_location_sharing else "Location Tracking"
        )
        for folder in self._gis.users.me.folders:
            if folder["title"] == folder_title:
                break
        else:
            self._gis.content.create_folder(folder_title)
        service_name = "location_tracking"
        if not self._gis.content.is_service_name_available(
            service_name, service_type="featureService"
        ):
            service_name = f"{service_name}{int(_dt.datetime.now().timestamp())}"
        create_params = {
            "name": f"{service_name}",
            "layers": [
                {
                    "adminLayerInfo": {
                        "tableMetadata": {
                            "numberOfShards": f"{tracks_layer_shards}",
                            "rollingIndexStrategy": f"{tracks_layer_rolling_index_strategy}",
                            "dataRetentionStrategy": "30",
                            "dataRetentionStrategyUnits": "DAYS",
                            "dataRetention": "true",
                        }
                    }
                },
                {
                    "adminLayerInfo": {
                        "tableMetadata": {
                            "numberOfShards": f"{lkl_layer_shards}",
                            "rollingIndexStrategy": "Yearly",
                            "dataRetentionStrategy": "30",
                            "dataRetentionStrategyUnits": "DAYS",
                            "dataRetention": "false",
                        }
                    }
                },
            ],
            "description": f"Location {'Sharing' if self._use_location_sharing else 'Tracking'} Service",
            "snippet": f"Location {'Sharing' if self._use_location_sharing else 'Tracking'} Service",
        }
        # Use a longer rolling index strategy if 10.8.1 or later
        if float(self._gis.properties.get("currentVersion", "0")) > 7.3:
            create_params["layers"][1]["adminLayerInfo"]["tableMetadata"][
                "rollingIndexStrategy"
            ] = "Century"
        item = self._gis.content.create_service(
            "location_tracking",
            create_params=create_params,
            folder=folder_title,
            service_type="locationTrackingService",
        )
        item.protect(True)
        item.update(
            {
                "description": f"The location {'Sharing' if self._use_location_sharing else 'Tracking'} service stores the last known location of each mobile user, "
                "as well as full historical tracks of where the mobile user has been. It is part of an "
                "organization-wide capability that is managed by an administrator.",
                "snippet": f"Location {'Sharing' if self._use_location_sharing else 'Tracking'} Service",
            }
        )
        item.share(org=True)
        self._gis.update_properties(
            {"locationTrackingService": {"url": item.url, "id": item.itemid}}
        )
        return True

    def pause(self):
        """
        Pauses location sharing for the organization.

        :return: True if successful, False otherwise
        """
        self._validate_environment()
        if self.status == "enabled":
            return bool(
                self._service.manager.update_definition({"capabilities": "Query"})[
                    "success"
                ]
            )
        else:
            return False

    def resume(self):
        """
        Resumes location sharing for the organization after it was paused.

        :return: True if successful, False otherwise
        """
        self._validate_environment()
        if self.status == "paused":
            return bool(
                self._service.manager.update_definition(
                    {"capabilities": "Query,Create,Update"}
                )["success"]
            )
        else:
            return False

    def disable(self):
        """
        Disables location sharing for the organization.

        THIS WILL DELETE ALL LOCATION SHARING VIEWS, SERVICES, AND DATA.

        :return: True if successful, False otherwise
        """
        item = self.item
        if self.status == "disabled":
            possible_lts = self._gis.content.search(
                query='typekeywords:"Location Tracking Service" NOT typekeywords:"Location Tracking View"'
            )
            if len(possible_lts) == 0:
                return False
            else:
                item = possible_lts[0]
        folder_id = item.ownerFolder
        views = item.related_items("Service2Service", "forward")
        for view in views:
            if (
                hasattr(view, "properties")
                and view.properties is not None
                and "trackViewGroup" in view.properties
            ):
                group = self._gis.groups.get(view.properties["trackViewGroup"])
                if group:
                    group.protected = False
                    group.delete()
            view.protect(False)
            view.delete()
        item.protect(False)
        item.delete()
        self._gis.update_properties({"locationTrackingService": "null"})
        if (
            len(
                self._gis.content.search(
                    """owner:"{}" ownerfolder:{}""".format(item.owner, folder_id)
                )
            )
            == 0
        ):
            folder_title = (
                "Location Sharing"
                if self._use_location_sharing
                else "Location Tracking"
            )
            for folder in self._gis.users.get(item.owner).folders:
                if folder["title"] == folder_title:
                    self._gis.content.delete_folder(folder["title"], owner=item.owner)
                    break
        return True

    def create_track_view(self, title: str):
        """
        This creates a :class:`~arcgis.apps.tracker.TrackView`.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        title                  Required String. The title of the Track View to create.
        ==================     ====================================================================

        :return: :class:`~arcgis.apps.tracker.TrackView`
        """
        self._validate_environment()
        folder = None
        if self.item.ownerFolder is not None:
            for f in self._gis.users.get(self.item.owner).folders:
                if f["id"] == self.item.ownerFolder:
                    folder = f["title"]
                    break
        group = self._gis.groups.create(
            title,
            f"Location {'Sharing' if self._use_location_sharing else 'Tracking'} Group",
            is_view_only=True,
            is_invitation_only=True,
            access="private",
        )
        group.protected = True
        item = self._gis.content.create_service(
            "{}_Track_View".format(group.id),
            create_params={"name": "{}_Track_View".format(group.id), "isView": True},
            folder=folder,
            service_type="locationTrackingService",
            is_view=True,
            owner=self.item.owner,
        )
        item.protect(True)
        item.update({"title": title, "properties": {"trackViewGroup": group.id}})
        definition = {"viewDefinitionQuery": "created_user in ('')"}
        # Workaround for 10.7 bug where time wasn't enabled
        if self._gis.properties.isPortal and self._gis.version <= [7, 1]:
            definition["timeInfo"] = {"startTimeField": "location_timestamp"}
        item.layers[0].manager.update_definition(definition)
        item.layers[1].manager.update_definition(definition)
        # Older versions of the LTS may not have this layer
        if len(item.layers) >= 3:
            item.layers[2].manager.update_definition(definition)
        if self._gis.properties.isPortal:
            # Set allowOthersToQuery to True - workaround for missing feature in Enterprise 10.7/10.7.1
            arcgis.features.FeatureLayerCollection(
                item.url, self._gis
            ).manager.update_definition(
                {
                    "editorTrackingInfo": {
                        "enableOwnershipAccessControl": True,
                        "enableEditorTracking": True,
                        "allowOthersToQuery": True,
                        "allowOthersToUpdate": False,
                        "allowOthersToDelete": False,
                    }
                }
            )
        item.share(groups=[group])
        if group.owner != self.item.owner:
            group.reassign_to(self.item.owner)
            group.remove_users([self._gis.users.me])
        return arcgis.apps.tracker.TrackView(item)

    @property
    def retention_period(self):
        """
        The retention period of the Location Sharing Tracks Layer.
        This is a positive integer whose units are defined by :attr:`~arcgis.apps.tracker.LocationTrackingManager.retention_period_units`
        """
        try:
            return int(
                self.tracks_layer.manager.properties["adminLayerInfo"]["tableMetadata"][
                    "dataRetentionStrategy"
                ]
            )
        except:
            return None

    @retention_period.setter
    def retention_period(self, value):
        self._validate_environment()
        if isinstance(value, str):
            if not value.isdigit():
                raise LocationTrackingError(
                    "Invalid Retention Policy Setting: '{}' expected an integer greater than 0".format(
                        value
                    )
                )
        elif isinstance(value, int):
            if value < 0:
                raise LocationTrackingError(
                    "Invalid Retention Policy Setting: '{}' expected an integer greater than 0".format(
                        value
                    )
                )
        else:
            raise LocationTrackingError(
                "Invalid Retention Policy Setting: '{}' expected an integer greater than 0".format(
                    value
                )
            )
        self.tracks_layer.manager.update_definition(
            {"tableMetadata": {"dataRetentionStrategy": "{}".format(value)}}
        )

    @property
    def retention_period_units(self):
        """The retention period units ("HOURS", "DAYS", "MONTHS", "YEARS") of the Location Sharing Tracks Layer"""
        try:
            return self.tracks_layer.manager.properties["adminLayerInfo"][
                "tableMetadata"
            ]["dataRetentionStrategyUnits"]
        except:
            return None

    @retention_period_units.setter
    def retention_period_units(self, value):
        self._validate_environment()
        if not isinstance(value, str):
            raise LocationTrackingError(
                "Invalid Retention Policy Setting: '{}' expected HOURS, DAYS, MONTHS, or YEARS".format(
                    value
                )
            )
        elif value.upper() not in ("HOURS", "DAYS", "MONTHS", "YEARS"):
            raise LocationTrackingError(
                "Invalid Retention Policy Setting: '{}' expected HOURS, DAYS, MONTHS, or YEARS".format(
                    value
                )
            )
        self.tracks_layer.manager.update_definition(
            {"tableMetadata": {"dataRetentionStrategyUnits": "{}".format(value)}}
        )

    @property
    def retention_period_enabled(self):
        """A boolean indicating if the retention period is enabled"""
        try:
            return (
                self.tracks_layer.manager.properties["adminLayerInfo"]["tableMetadata"][
                    "dataRetention"
                ].lower()
                == "true"
            )
        except:
            return None

    @retention_period_enabled.setter
    def retention_period_enabled(self, value):
        self._validate_environment()
        if isinstance(value, str):
            if value.lower() not in ("false", "true"):
                raise LocationTrackingError(
                    "Invalid Retention Policy Setting: '{}' expected True or False".format(
                        value
                    )
                )
        elif not isinstance(value, bool):
            raise LocationTrackingError(
                "Invalid Retention Policy Setting: '{}' expected True or False".format(
                    value
                )
            )
        self.tracks_layer.manager.update_definition(
            {"tableMetadata": {"dataRetention": "{}".format(str(value).lower())}}
        )

    @_lazy_property
    def item(self):
        """The Location Sharing :class:`~arcgis.gis.Item`"""
        try:
            return self._gis.content.get(
                self._gis.properties.helperServices["locationTracking"]["id"]
            )
        except:
            return None

    @_lazy_property
    def tracks_layer(self):
        """The tracks :class:`~arcgis.features.FeatureLayer`"""
        try:
            return self.item.layers[0]
        except:
            return None

    @_lazy_property
    def last_known_locations_layer(self):
        """The last known locations :class:`~arcgis.features.FeatureLayer`"""
        try:
            return self.item.layers[1]
        except:
            return None

    @_lazy_property
    def track_lines_layer(self):
        """The track lines :class:`~arcgis.features.FeatureLayer`"""
        try:
            return self.item.layers[2]
        except IndexError:
            return None

    @property
    def status(self):
        """The status of location sharing ("disabled", "paused", "enabled")"""
        try:
            if "locationTracking" not in self._gis.properties.helperServices:
                return "disabled"
            if "Create" in self._service.properties["capabilities"]:
                return "enabled"
            else:
                return "paused"
        except:
            return "disabled"

    @_lazy_property
    def _service(self):
        try:
            return arcgis.features.FeatureLayerCollection(
                self._gis.properties.helperServices["locationTracking"]["url"],
                self._gis,
            )
        except:
            return None



from __future__ import annotations
from arcgis.gis.kubernetes._admin._base import _BaseKube
from arcgis.gis import GIS


###########################################################################
class WebAdaptorManager(_BaseKube):
    """
    Provides access to the web adaptor resources defined on the ArcGIS
    Enterprise.
    """

    _gis = None
    _con = None
    _properties = None
    _url = None

    def __init__(self, url: str, gis: GIS):
        super()
        self._url = url
        self._gis = gis
        self._con = gis._con

    # ---------------------------------------------------------------------
    def web_adaptor(self, adaptor_id: str):
        """
        This resource returns the properties of an individual web adaptor, such as the HTTP and HTTPS ports.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        adaptor_id             Required string. The specific web adaptor to get.
        ==================     ====================================================================
        """
        url = f"{self._url}/{adaptor_id}"
        params = {"f": "json"}
        return self._con.get(url, params)

    # ---------------------------------------------------------------------
    def unregister_adaptor(self, adaptor_id: str):
        """
        This operation unregisters an ArcGIS Enterprise on Kubernetes Web Adaptor
        from your deployment. Once a web adaptor has been unregistered, the
        web adaptor will no longer be trusted and its credentials will not be
        accepted. This operation is typically used when you want to register
        a new ArcGIS Enterprise on Kubernetes Web Adaptor or when the previous
        one needs to be updated.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        adaptor_id             Required string. The web adaptor to unregister.
        ==================     ====================================================================
        """
        url = f"{self._url}/{adaptor_id}/unregister"
        params = {"f": "json"}
        return self._con.post(url, params)

    # ---------------------------------------------------------------------
    @property
    def configuration(self):
        """
        This resource is a collection of configuration properties that apply
        to the ArcGIS Enterprise on Kubernetes Web Adaptor configured with your deployment.
        The only supported property is sharedKey, which represents credentials
        that are shared with the web adaptor. The web adaptor will use these credentials
        to communicate with your deployment.
        """
        url = f"{self._url}/config"
        params = {"f": "json"}
        return self._con.get(url, params)

    # ---------------------------------------------------------------------
    @configuration.setter
    def configuration(self, configs: dict):
        """
        This operation is used to change the common properties and configurations
        for the ArcGIS Enterprise on Kubernetes Web Adaptor configured with your
        deployment. The properties are stored in a JSON object. Therefore,
        every update must include all necessary properties.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        configs                Required dictionary. The new configs for the web adaptors
        ==================     ====================================================================
        """
        url = f"{self._url}/config/update"
        params = {"f": "json", "webAdaptorsConfig": configs}
        return self._con.post(url, params)



"""
This is the ArcGIS Notebook Server API Framework
"""

import os
import copy
import warnings
from urllib.parse import urlparse

from arcgis.gis import GIS
from arcgis._impl.common._mixins import PropertyMap

from ._logs import LogManager
from ._system import SystemManager
from ._security import SecurityManager
from ._machines import MachineManager
from ._nbm import NotebookManager


########################################################################
class NotebookServer(object):
    """
    Provides access to the ArcGIS Notebook Server administration API.
    """

    _da = None
    _gis = None
    _url = None
    _properties = None
    _logs = None
    _system = None
    _machine = None
    _notebook = None
    _security = None
    _services = None
    _version = None
    _sitemanager = None

    # ----------------------------------------------------------------------
    def __init__(self, url, gis):
        """Constructor"""
        if url.lower().endswith("/admin") == False:
            url += "/admin"
        self._url = url
        if isinstance(gis, GIS):
            self._gis = gis
        else:
            raise ValueError("Invalid GIS object")

    # ----------------------------------------------------------------------
    def _init(self):
        """loads the properties"""
        try:
            params = {"f": "json"}
            res = self._gis._con.get(self._url, params)
            self._properties = PropertyMap(res)
        except:
            self._properties = PropertyMap({})

    # ----------------------------------------------------------------------
    def __str__(self):
        return "< NotebookServer @ {url} >".format(url=self._url)

    # ----------------------------------------------------------------------
    def __repr__(self):
        return "< NotebookServer @ {url} >".format(url=self._url)

    # ----------------------------------------------------------------------
    @property
    def properties(self):
        """Properties of the object"""
        if self._properties is None:
            self._init()
        return self._properties

    # ----------------------------------------------------------------------
    @property
    def data_access(self) -> "NotebookDataAccess":
        """Provides access to managing files stored on notebook server.

        :return:
            :class:`~arcgis.gis.nb._dataaccess.NotebookDataAccess` object

        """
        if self._da is None:
            from ._dataaccess import NotebookDataAccess

            url = self._url + "/dataaccess"
            self._da = NotebookDataAccess(url, self._gis)
        return self._da

    # ----------------------------------------------------------------------
    @property
    def version(self):
        """
        Returns the notebook server version

        :return: List
        """
        if self._version is None:
            self._version = [int(i) for i in self.properties.version.split(".")]
        return self._version

    # ----------------------------------------------------------------------
    @property
    def site(self):
        """
        Provides access to the notebook server's site management operations

        :return: :class:`~arcgis.gis.nb.SiteManager`
        """
        if self._sitemanager is None:
            from ._site import SiteManager

            self._sitemanager = SiteManager(url=self._url, notebook=self, gis=self._gis)
        return self._sitemanager

    # ----------------------------------------------------------------------
    @property
    def info(self):
        """
        Returns information about the server site itself

        :return: PropertyMap

        """
        url = self._url + "/info"
        params = {"f": "json"}
        res = self._gis._con.get(url, params)
        return PropertyMap(res)

    # ----------------------------------------------------------------------
    @property
    def health_check(self) -> bool:
        """

        The `health_check` verifies that your ArcGIS Notebook Server site
        has been created, and that its Docker environment has been
        correctly configured.

        **This is only avaible if the site can be accessed around the web adapter**

        :return: Boolean

        """
        netloc = urlparse(self._url).netloc
        if netloc.find(":11443") > -1:
            url = "https://{base}/arcgis/rest/info/healthcheck".format(base=netloc)
        else:
            url = "https://{base}:11443/arcgis/rest/info/healthcheck".format(
                base=netloc
            )
        params = {"f": "json"}
        verify_original = copy.deepcopy(self._gis._con._verify_cert)
        if self._gis._con._verify_cert:
            self._gis._con._verify_cert = False
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            res = self._gis._con.get(url, params)
        self._gis._con._verify_cert = verify_original
        if "success" in res:
            return res["success"]
        return res

    # ----------------------------------------------------------------------
    @property
    def logs(self):
        """
        Provides access to the notebook server's logging system

        :return:
            :class:`~arcgis.gis.nb.LogManager`

        """
        if self._logs is None:
            url = self._url + "/logs"
            self._logs = LogManager(url=url, gis=self._gis)
        return self._logs

    # ----------------------------------------------------------------------
    @property
    def system(self):
        """
        returns access to the system properties of the ArcGIS Notebook Server

        :return:
            :class:`~arcgis.gis.nb.SystemManager`

        """
        if self._system is None:
            url = self._url + "/system"
            self._system = SystemManager(url=url, gis=self._gis)
        return self._system

    # ----------------------------------------------------------------------
    @property
    def machine(self):
        """
        Provides access to managing the registered machines with ArcGIS
        Notebook Server

        :return: :class:`~arcgis.gis.nb.MachineManager`

        """
        if self._machine is None:
            url = self._url + "/machines"
            self._machine = MachineManager(url=url, gis=self._gis)
        return self._machine

    # ----------------------------------------------------------------------
    @property
    def security(self):
        """
        Provides access to managing the ArcGIS Notebook Server's security
        settings.

        :return: :class:`~arcgis.gis.nb.SecurityManager`

        """
        if self._security is None:
            url = self._url + "/security"
            self._security = SecurityManager(url=url, gis=self._gis)
        return self._security

    # ----------------------------------------------------------------------
    @property
    def notebooks(self):
        """
        Provides access to managing the ArcGIS Notebook Server's
        Notebooks

        :return: :class:`~arcgis.gis.nb.NotebookManager`
        """
        if self._notebook is None:
            url = self._url + "/notebooks"
            self._notebook = NotebookManager(url=url, gis=self._gis, nbs=self)
        return self._notebook

    # ----------------------------------------------------------------------
    @property
    def url(self):
        """The URL of the notebook server."""
        return self._url

    # ----------------------------------------------------------------------
    @property
    def services(self):
        """
        Provices access to managing notebook created geoprocessing tools


        :return:
            :class:`~arcgis.gis.nb._services.NBServicesManager`

        """
        if self._services is None:
            from arcgis.gis.nb._services import NBServicesManager

            url = self._url + "/services"
            self._services = NBServicesManager(url, self._gis, nbs=self)
        return self._services



from .notebook import NotebookServer
from ._nbm import NotebookManager, Notebook, Runtime
from ._snapshot import SnapshotManager, SnapShot
from ._system import (
    SystemManager,
    Container,
    ContainerNotebook,
    DirectoryManager,
    WebAdaptor,
    WebAdaptorManager,
)

from ._site import SiteManager
from ._logs import LogManager
from ._machines import MachineManager, Machine
from ._security import SecurityManager


__all__ = [
    "NotebookServer",
    "NotebookManager",
    "Notebook",
    "Runtime",
    "SnapshotManager",
    "SnapShot",
    "SystemManager",
    "Container",
    "ContainerNotebook",
    "DirectoryManager",
    "WebAdaptor",
    "WebAdaptorManager",
    "SiteManager",
    "LogManager",
    "MachineManager",
    "Machine",
    "SecurityManager",
]



from typing import Any, Optional, Union, Dict
from arcgis.gis import GIS
from arcgis.raster import _util


def generate_direct_access_url(
    expiration: Optional[int] = None, *, gis: Optional[GIS] = None
):
    """
    Function to get the direct access url for user's rasterStore on ArcGIS Online.

    ====================================     ====================================================================
    **Parameter**                             **Description**
    ------------------------------------     --------------------------------------------------------------------
    expiration                               Optional integer. Direct access URL expiration time in minutes.
                                             (The default is 1440 ie. 24 hours)
    ------------------------------------     --------------------------------------------------------------------
    gis                                      Keyword only parameter. Optional :class:`~arcgis.gis.GIS` . The GIS on which this function runs.
                                             If not specified, the active GIS is used.
    ====================================     ====================================================================

    :return:
        String. Direct Access URL
    """

    return _util._generate_direct_access_url(expiration=expiration, gis=gis)


def upload_imagery_to_agol_userstore(
    files: Union[str, list],
    direct_access_url: Optional[str] = None,
    auto_renew: bool = True,
    upload_properties: Optional[dict] = None,
    *,
    gis: Optional[GIS] = None,
):
    """
    Uploads file/files to the user's rasterstore on ArcGIS Online and returns the list of urls.
    
    The list of urls can then be used with :meth:`~arcgis.raster.analytics.copy_raster` or :meth:`~arcgis.raster.analytics.create_image_collection`
    method to create imagery layers on ArcGIS Online.
    
    For this functionality to work, Azure library packages for Python (Azure SDK for Python - azure-storage-blob: 12.1<= version <=12.17)
    needs to be pre-installed. Refer https://docs.microsoft.com/en-us/azure/storage/blobs/storage-quickstart-blobs-python#install-the-package

    ====================================     ====================================================================
    **Parameter**                             **Description**
    ------------------------------------     --------------------------------------------------------------------
    files                                    Required. It can be a folder, list of files or single file that needs to be uploaded.
    ------------------------------------     --------------------------------------------------------------------
    direct_access_url                        Optional string. The direct access url generated using :meth:`~arcgis.raster.utils.generate_direct_access_url` .
                                             If not specified, the function would generate the direct access url internally which is valid for 1440 minutes.
    ------------------------------------     --------------------------------------------------------------------
    auto_renew                               Optional boolean. If set to True, function would continue uploading 
                                             until the entire data is uploaded by auto renewing the direct access url.
                                             (The default is True)
    ------------------------------------     --------------------------------------------------------------------
    upload_properties                        | Optional dictionary. ``upload_properties`` can be used to control specific \
                                             upload parameters. 

                                             Available options:

                                             - ``maxUploadConcurrency``: Optional integer. Maximum number of parallel connections \
                                                to use for large uploads (when individual file/blob size exceeds 64MB). \
                                                This is the **max_concurrency** parameter of the `BlobClient.upload_blob() <https://docs.microsoft.com/en-us/python/api/azure-storage-blob/azure.storage.blob.blobclient?view=azure-python#upload-blob-data--blob-type--blobtype-blockblob---blockblob----length-none--metadata-none----kwargs->`__ method. \
                                                (The default is 6)
                                             - ``maxWorkerThreads``: Optional integer. Maximum number of threads to execute asynchronously \
                                                when uploading multiple files. This is the **max_workers** parameter of the `ThreadPoolExecutor() <https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ThreadPoolExecutor>`__ class. \
                                                (The default is None)
                                             - ``displayProgress``: Optional boolean. If set to True, a progress bar will be \
                                                displayed for tracking the progress of the uploads to user's rasterstore. \
                                                (The default is False)

                                               Example:

                                                    | {"maxUploadConcurrency":8,
                                                    | "maxWorkerThreads":20,
                                                    | "displayProgress":True}
    ------------------------------------     --------------------------------------------------------------------
    gis                                      Keyword only parameter. Optional :class:`~arcgis.gis.GIS` . The GIS on which this function runs.
                                             If not specified, the active GIS is used.
    ====================================     ====================================================================

    :return:
        List of file paths.

    .. code-block:: python

        # Usage Example: Generates an expirable direct access url and uploads files to the user's raster store.

        sas_url = generate_direct_access_url(expiration=180, gis=gis)

        uploaded_imagery = upload_imagery_to_agol_userstore(files=r"/path/to/data", 
                                                            direct_access_url=sas_url,
                                                            upload_properties={"displayProgress":True},
                                                            gis=gis
                                                            )

        # Following snippet executes the copy_raster() function on the uploaded imagery to create imagery layer item on ArcGIS Online.

        copy_raster_op = copy_raster(input_raster=uploaded_imagery,
                                     raster_type_name="Raster Dataset",
                                     output_name="output_layer",
                                     gis=gis)

        """

    return _util._upload_imagery_agol(
        files=files,
        direct_access_url=direct_access_url,
        auto_renew=auto_renew,
        upload_properties=upload_properties,
        gis=gis,
    )


def publish_hosted_imagery_layer(
    input_data: list,
    layer_configuration: str,
    tiles_only: Optional[bool] = False,
    raster_type_name: Optional[str] = None,
    raster_type_params: Optional[Dict[str, Any]] = None,
    source_mosaic: Optional[str] = None,
    output_name: Optional[str] = None,
    context: Optional[Dict[str, Any]] = None,
    *,
    gis: Optional[GIS] = None,
    future: bool = False,
    **kwargs,
):
    """
    The function can create hosted imagery layers in ArcGIS Enterprise and ArcGIS Online 
    from local raster datasets by uploading the data to the server. 
    Multiple images are mosaicked into a single dataset to create one layer. 
    A collection can also be created from multiple input rasters.

    ====================================     ====================================================================
    **Parameter**                             **Description**
    ------------------------------------     --------------------------------------------------------------------
    input_data                               Required list. The list of input raster paths(s) to be added to 
                                             the imagery layer being created. 
                                             
                                             .. note:: 
                                                 You can also upload an existing mosaic dataset and create an imagery layer by 
                                                 specifying the mosaic dataset path as input to source_mosaic parameter. 
                                                 The input_data param can then be used to specify local raster dataset path(s) in the mosaic dataset.
    ------------------------------------     --------------------------------------------------------------------
    layer_configuration                      Required String.

                                                - ONE_IMAGE: Uses a single, processed image or mosaics multiple \
                                                images into a single dataset to create one layer. \
                                                This option supports all common image formats and satellite \
                                                products at various levels.

                                                - IMAGE_COLLECTION: Manages a collection of images using a \
                                                single layer and mosaics images dynamically. Each image can \
                                                be accessed independently. This option supports all common \
                                                image formats and satellite products at various levels. 
    ------------------------------------     --------------------------------------------------------------------
    tiles_only                               Optional boolean. If set to True, function will generate a tiles only layer
                                             otherwise will generate a dynamic imagery layer.

                                                - True - Provides imagery access as static tiles and associated metadata. \
                                                Supports client-side processing and rendering. Can be used as an input \
                                                to raster analysis.

                                                - False - Provides versatile dynamic imagery access capabilities. \
                                                Supports on-demand server-side processing and \
                                                dynamic mosaicking. Supports managing a collection of images. \
                                                Can be used as an input to raster analysis.
    ------------------------------------     --------------------------------------------------------------------
    raster_type_name                         Required string. The name of the raster type to use for adding data.

                                             Choice list:

                                                | [
                                                | "Aerial", "ASTER", "DMCII", "DubaiSat-2", "GeoEye-1", "GF-1 PMS", "GF-1 WFV",
                                                | "GF-2 PMS", "GRIB", "HDF", "IKONOS", "Jilin-1", "KOMPSAT-2", "KOMPSAT-3",
                                                | "Landsat 1-5 MSS", "Landsat 4-5 TM", "Landsat 7 ETM+", "Landsat 8", "Landsat 9",
                                                | "NetCDF", "PlanetScope", "Pleiades-1", "Pleiades NEO", "QuickBird", "RapidEye",
                                                | "Raster Dataset", "ScannedAerial", "Sentinel-2", "SkySat", "SPOT 5", "SPOT 6",
                                                | "SPOT 7", "Superview-1", "Tiled Imagery Layer", "UAV/UAS", "WordView-1",
                                                | "WordView-2", "WordView-3", "WordView-4", "ZY3-SASMAC", "ZY3-CRESDA"
                                                | ]

                                             If an existing mosaic dataset is being published as a 
                                             dynamic imagery layer using the ``source_mosaic_dataset`` parameter, the
                                             ``raster_type_name`` parameter can be set to None as it is not required.


                                             Example:

                                                "QuickBird"
    ------------------------------------     --------------------------------------------------------------------
    raster_type_params                       Optional dict. Additional ``raster_type`` specific parameters.
        
                                             The process of add rasters can be controlled by specifying \
                                             additional raster type arguments.

                                             The raster type parameters argument is a dictionary.

                                             The dictionary can contain productType, processingTemplate, \
                                             pansharpenType, Filter, pansharpenWeights, ConstantZ, \
                                             dem, zoffset, CorrectGeoid, ZFactor, StretchType, \
                                             ScaleFactor, ValidRange

                                             Please check the table below (Supported Raster Types), \
                                             for more details about the product types, \
                                             processing templates, pansharpen weights for each raster type. 

                                             - Possible values for pansharpenType - ["Mean", "IHS", "Brovey", "Esri", "Mean", "Gram-Schmidt"]
                                             - Possible values for filter - [None, "Sharpen", "SharpenMore"]
                                             - Value for StretchType dictionary can be as follows:

                                               - "None"
                                               - "MinMax; <min>; <max>"
                                               - "PercentMinMax; <MinPercent>; <MaxPercent>"
                                               - "StdDev; <NumberOfStandardDeviation>"
                                               Example: {"StretchType": "MinMax; <min>; <max>"}
                                             - Value for ValidRange dictionary can be as follows:

                                               - "<MaskMinValue>, <MaskMaxValue>"
                                               Example: {"ValidRange": "10, 200"}

                                             Example:

                                                {"productType":"All","processingTemplate":"Pansharpen",
                                                "pansharpenType":"Gram-Schmidt","filter":"SharpenMore",
                                                "pansharpenWeights":"0.85 0.7 0.35 1","constantZ":-9999}
    ------------------------------------     --------------------------------------------------------------------
    source_mosaic_dataset                    Optional string. Path to the existing mosaic dataset to be published 
                                             as a hosted dynamic imagery layer.

                                             To publish an existing mosaic dataset, specify the path to the input 
                                             data of the mosaic in the ``input_rasters`` parameter. 
                                             The data will be uploaded to ArcGIS Online.

                                             ``raster_type_name`` parameter can be set to None as it is not required to 
                                             publish an imagery layer from a mosaic dataset.

                                             .. note::
                                                    Option available only on ArcGIS online

                                             Example:

                                                "./data/temp_uploaded.gdb/test"
    ------------------------------------     --------------------------------------------------------------------
    output_name                              Optional String. If not provided, an Image Service is created by the method 
                                             and used as the output raster.

                                             You can pass in an existing Image Service Item from your GIS to use 
                                             that instead.

                                             Alternatively, you can pass in the name of the output Image Service 
                                             that should be created by this method to be used as the output for 
                                             the tool.
                                             A RuntimeError is raised if a service by that name already exists
    ------------------------------------     --------------------------------------------------------------------
    context                                  context contains additional settings that affect task execution.

                                             context parameter overwrites values set through arcgis.env parameter
                                         
                                             This function has the following settings:

                                             - Output Spatial Reference (outSR): The output raster will be projected into the output spatial reference.
                                                
                                               Example:

                                                    {"outSR": {spatial reference}}


                                             - Upload Properties (upload_properties): ``upload_properties`` key can be used to control specific upload parameters when trying to create hosted imagery layers in ArcGIS Online from local raster datasets.

                                               Available options:

                                               - ``maxUploadConcurrency``: Optional integer. Maximum number of parallel connections \
                                                 to use for large uploads (when individual file/blob size exceeds 64MB). \
                                                 This is the **max_concurrency** parameter of the `BlobClient.upload_blob() <https://docs.microsoft.com/en-us/python/api/azure-storage-blob/azure.storage.blob.blobclient?view=azure-python#upload-blob-data--blob-type--blobtype-blockblob---blockblob----length-none--metadata-none----kwargs->`__ method. \
                                                 (The default is 6)
                                               - ``maxWorkerThreads``: Optional integer. Maximum number of threads to execute asynchronously \
                                                 when uploading multiple files. This is the **max_workers** parameter of the `ThreadPoolExecutor() <https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ThreadPoolExecutor>`__ class. \
                                                 (The default is None)
                                               - ``displayProgress``: Optional boolean. If set to True, a progress bar will be \
                                                 displayed for tracking the progress of the uploads to user's rasterstore. \
                                                 (The default is False)

                                               Example:

                                                    | {"upload_properties": {"maxUploadConcurrency":8,
                                                    |                       "maxWorkerThreads":20,
                                                    |                       "displayProgress":True}
                                                    | }

                                             The context parameter can also be used to specify whether to
                                             build footprints, pixel value that represents the NoData,
                                             resamplingMethod etc.


                                             Example:

                                                | {"buildFootprints":True,                                            
                                                | "footprintsArguments":{"method":"RADIOMETRY","minValue":1,"maxValue":5,
                                                |                        "shrinkDistance":50,"skipOverviews":True,"updateBoundary":True,
                                                |                        "maintainEdge":False,"simplification":None,"numVertices":20,
                                                |                        "minThinnessRatio":0.05,"maxSliverSize":20,"requestSize":2000,
                                                |                        "minRegionSize":100},
                                                | "defineNodata":True,                                            
                                                | "noDataArguments":{"noDataValues":[500],"numberOfBand":99,"compositeValue":True},                                            
                                                | "buildOverview":True}
    ------------------------------------     --------------------------------------------------------------------
    gis                                      Keyword only parameter. Optional :class:`~arcgis.gis.GIS` . The GIS on which this function runs.
                                             If not specified, the active GIS is used.
    ====================================     ====================================================================

    :return:
    Imagery layer item

        """
    from arcgis import env

    if output_name is None:
        from arcgis.raster._util import _id_generator

        output_name = "layer" + "_" + _id_generator()

    if layer_configuration.upper() == "ONE_IMAGE":
        gis = env.active_gis if gis is None else gis

        return gis._tools.rasteranalysis.copy_raster(
            input_raster=input_data,
            output_name=output_name,
            context=context,
            future=future,
            raster_type_name=raster_type_name,
            raster_type_params=raster_type_params,
            md_to_upload=source_mosaic,
            tiles_only=tiles_only,
            **kwargs,
        )

    elif layer_configuration.upper() == "IMAGE_COLLECTION":
        gis = env.active_gis if gis is None else gis

        return gis._tools.rasteranalysis.create_image_collection(
            image_collection=output_name,
            input_rasters=input_data,
            raster_type_name=raster_type_name,
            raster_type_params=raster_type_params,
            md_to_upload=source_mosaic,
            context=context,
            gis=gis,
            future=future,
            **kwargs,
        )

    else:
        raise RuntimeError(
            "layer_configuration should be either ONE_IMAGE or IMAGE_COLLECTION"
        )



from __future__ import annotations
from ._execute import execute_notebook, list_runtimes
from ._snapshots import list_snapshots, create_snapshot

__all__ = ["execute_notebook", "list_runtimes", "list_snapshots", "create_snapshot"]



from ._arcgis_model import ArcGISModel
import traceback
import json, time, datetime
from datetime import datetime as dt
import pandas as pd
from IPython.display import clear_output

try:
    import arcgis as ag
    import torch, sys
    from . import MMSegmentation, MMDetection
    import matplotlib.pyplot as plt
    from ._autodl_utils import (
        train_callback,
        ToolIsCancelled,
        _train_exhaust_mode,
        _get_emd_path,
    )
    from .._utils.evaluate_batchsize import estimate_batch_size
    from ._arcgis_model import ArcGISModel
    from .._data import prepare_data
    import numpy as np
    import time, gc
    import optuna, os
    from io import StringIO
    import os, json
except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False

self_obj = None


class RedirectedStdout:
    def __init__(self):
        self._stdout = None
        self._string_io = None

    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._string_io = StringIO()
        return self

    def __exit__(self, type, value, traceback):
        sys.stdout = self._stdout

    def __str__(self):
        return self._string_io.getvalue()


class ImageryModel(ArcGISModel):
    """
    Imagery Model is used to fine tune models trained using AutoDL
    """

    def __init__(self):
        self._modeltype = None
        pass

    def load(self, path, data):
        """
        Loads a compatible saved model for inferencing or fine tuning from the disk,
        which can be used to further fine tune the models saved using AutoDL.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        path                    Required string. Path to
                                Esri Model Definition(EMD) or DLPK file.
        ---------------------   -------------------------------------------
        data                    Required ImageryDataObject. Returned data
                                object from :meth:`~arcgis.learn.prepare_data`  function.
        =====================   ===========================================
        """
        is_mm = False
        try:
            emd_path = _get_emd_path(path)
            f = open(emd_path)
            emd = json.load(f)
            f.close()
        except json.decoder.JSONDecodeError as E:
            raise Exception(E)
        except Exception as e:
            print(e)
            raise Exception("Not a supported Esri Model Definition(EMD) or DLPK file")
        try:
            modelname = emd["ModelName"]
            self._modeltype = emd["ModelType"]
            if "ModelFileConfigurationClass" in list(emd.keys()):
                self._modelconfig = emd["ModelFileConfigurationClass"]
                if self._modelconfig in ["MMDetectionConfig", "MMSegmentationConfig"]:
                    mm_model = emd["Kwargs"]["model"]
                    is_mm = True
        except Exception as e:
            print(e)
            raise Exception("Not a supported Esri Model Definition(EMD) or DLPK file")
        if is_mm:
            setattr(
                self,
                "imagery_model",
                getattr(ag.learn, modelname)(data, model=mm_model),
            )
        else:
            setattr(
                self,
                "imagery_model",
                getattr(ag.learn, modelname).from_model(path, data),
            )
        getattr(self, "imagery_model").load(path)

    def fit(
        self,
        epochs=10,
        lr=None,
        one_cycle=True,
        early_stopping=False,
        checkpoint=True,
        tensorboard=False,
        monitor="valid_loss",
        **kwargs
    ):
        """
        Train the model for the specified number of epochs while using the
        specified learning rates

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        epochs                  Optional integer. Number of cycles of training
                                on the data. Increase it if the model is underfitting.
                                Default value is 10.
        ---------------------   -------------------------------------------
        lr                      Optional float or slice of floats. Learning rate
                                to be used for training the model. If ``lr=None``,
                                an optimal learning rate is automatically deduced
                                for training the model.
        ---------------------   -------------------------------------------
        one_cycle               Optional boolean. Parameter to select 1cycle
                                learning rate schedule. If set to `False` no
                                learning rate schedule is used.
        ---------------------   -------------------------------------------
        early_stopping          Optional boolean. Parameter to add early stopping.
                                If set to 'True' training will stop if parameter
                                `monitor` value stops improving for 5 epochs.
                                A minimum difference of 0.001 is required for
                                it to be considered an improvement.
        ---------------------   -------------------------------------------
        checkpoint              Optional boolean or string.
                                Parameter to save checkpoint during training.
                                If set to `True` the best model
                                based on `monitor` will be saved during
                                training. If set to 'all', all checkpoints
                                are saved. If set to False, checkpointing will
                                be off. Setting this parameter loads the best
                                model at the end of training.
        ---------------------   -------------------------------------------
        tensorboard             Optional boolean. Parameter to write the training log.
                                If set to 'True' the log will be saved at
                                <dataset-path>/training_log which can be visualized in
                                tensorboard. Required tensorboardx version=2.1

                                The default value is 'False'.

                                .. note::

                                    Not applicable for Text Models
        ---------------------   -------------------------------------------
        monitor                 Optional string. Parameter specifies
                                which metric to monitor while checkpointing
                                and early stopping. Defaults to 'valid_loss'. Value
                                should be one of the metric that is displayed in
                                the training table. Use ``{model_name}.available_metrics``
                                to list the available metrics to set here.
        =====================   ===========================================
        """
        try:
            getattr(self, "imagery_model").fit(
                epochs,
                lr,
                one_cycle,
                early_stopping,
                checkpoint,
                tensorboard,
                monitor,
                **kwargs
            )
        except Exception as E:
            print("Load the model first using load()")

    def show_results(self, rows=5, **kwargs):
        """
        Displays the results of a trained model on a part of the validation set.

        =====================   ===========================================
        rows                     Optional int. Number of rows of results
                                 to be displayed.
        =====================   ===========================================
        """
        getattr(self, "imagery_model").show_results(rows, **kwargs)

    def save(
        self,
        name_or_path,
        framework="PyTorch",
        publish=False,
        gis=None,
        compute_metrics=True,
        save_optimizer=False,
        save_inference_file=True,
        **kwargs
    ):
        """
        Saves the model weights, creates an Esri Model Definition and Deep
        Learning Package zip for deployment to Image Server or ArcGIS Pro.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Name of the model to save. It
                                stores it at the pre-defined location. If path
                                is passed then it stores at the specified path
                                with model name as directory name and creates
                                all the intermediate directories.
        ---------------------   -------------------------------------------
        framework               Optional string. Exports the model in the
                                specified framework format ('PyTorch', 'tflite'
                                'torchscript', and 'TF-ONXX' (deprecated)).
                                Only models saved with the default framework
                                (PyTorch) can be loaded using `from_model`.
                                ``tflite`` framework (experimental support) is
                                supported by :class:`~arcgis.learn.SingleShotDetector`,
                                :class:`~arcgis.learn.FeatureClassifier` and  :class:`~arcgis.learn.RetinaNet` .
                                ``torchscript`` format is supported by
                                :class:`~arcgis.learn.SiamMask` .
                                For usage of SiamMask model in ArcGIS Pro 2.8,
                                load the ``PyTorch`` framework saved model
                                and export it with ``torchscript`` framework
                                using ArcGIS API for Python v1.8.5.
                                For usage of SiamMask model in ArcGIS Pro 2.9,
                                set framework to ``torchscript`` and use the
                                model files additionally generated inside
                                'torch_scripts' folder.
                                If framework is ``TF-ONNX`` (Only supported for
                                :class:`~arcgis.learn.SingleShotDetector`), ``batch_size`` can
                                be passed as an optional keyword argument.
        ---------------------   -------------------------------------------
        publish                 Optional boolean. Publishes the DLPK as an item.
        ---------------------   -------------------------------------------
        gis                     Optional :class:`~arcgis.gis.GIS`  Object. Used for publishing the item.
                                If not specified then active gis user is taken.
        ---------------------   -------------------------------------------
        compute_metrics         Optional boolean. Used for computing model
                                metrics.
        ---------------------   -------------------------------------------
        save_optimizer          Optional boolean. Used for saving the model-optimizer
                                state along with the model. Default is set to False
        ---------------------   -------------------------------------------
        save_inference_file     Optional boolean. Used for saving the inference file
                                along with the model.
                                If False, the model will not work with ArcGIS Pro 2.6
                                or earlier. Default is set to True.
        ---------------------   -------------------------------------------
        kwargs                  Optional Parameters:
                                Boolean `overwrite` if True, it will overwrite
                                the item on ArcGIS Online/Enterprise, default False.
        =====================   ===========================================
        """
        getattr(self, "imagery_model").save(
            name_or_path,
            framework,
            publish,
            gis,
            compute_metrics,
            save_optimizer,
            save_inference_file,
            **kwargs
        )

    def lr_find(self, allow_plot=True):
        """
        Runs the Learning Rate Finder. Helps in choosing the
        optimum learning rate for training the model.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        allow_plot              Optional boolean. Display the plot of losses
                                against the learning rates and mark the optimal
                                value of the learning rate on the plot.
                                The default value is 'True'.
        =====================   ===========================================
        """
        try:
            lr = getattr(self, "imagery_model").lr_find(allow_plot)
            return lr
        except Exception as E:
            print("Load the model first using load()")

    def available_metrics(self):
        """
        List of available metrics that are displayed in the training
        table. Set `monitor` value to be one of these while calling
        the :class:`~arcgis.learn.ImageryModel.fit` method.
        """
        try:
            return getattr(self, "imagery_model").available_metrics
        except Exception as E:
            print("Load the model first using load()")

    def plot_losses(self):
        """
        Plot validation and training losses after fitting the model.
        """
        try:
            return getattr(self, "imagery_model").plot_losses()
        except Exception as E:
            print("Load the model first using load()")

    def unfreeze(self):
        """
        Unfreezes the earlier layers of the model for fine-tuning.
        """
        try:
            if hasattr(getattr(self, "imagery_model"), "unfreeze"):
                return getattr(self, "imagery_model").unfreeze()
            else:
                print("This model does not support unfreeze method")
        except Exception as E:
            print("Load the model first using load()")

    def mIOU(self):
        """
        Computes mean IOU on the validation set for each class.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        mean                    Optional bool. If False returns class-wise
                                mean IOU, otherwise returns mean iou of all
                                classes combined.
        ---------------------   -------------------------------------------
        show_progress           Optional bool. Displays the progress bar if
                                True.
        =====================   ===========================================

        :return: `dict` if mean is False otherwise `float`
        """
        if self._modeltype is not None:
            if self._modeltype == "ObjectDetection":
                print("This method is not supported with Object Detection models")
                return
            else:
                try:
                    return getattr(self, "imagery_model").mIOU()
                except Exception as E:
                    print("Load the model first using load()")
        else:
            print("Train the model first using fit()")
            return

    def average_precision_score(self):
        """
        Computes average precision on the validation set for each class.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        detect_thresh           Optional float. The probability above which
                                a detection will be considered for computing
                                average precision.
        ---------------------   -------------------------------------------
        iou_thresh              Optional float. The intersection over union
                                threshold with the ground truth labels, above
                                which a predicted bounding box will be
                                considered a true positive.
        ---------------------   -------------------------------------------
        mean                    Optional bool. If False returns class-wise
                                average precision otherwise returns mean
                                average precision.
        =====================   ===========================================

        :return: `dict` if mean is False otherwise `float`
        """
        if self._modeltype is not None:
            if self._modeltype != "ObjectDetection":
                print("This method is not supported with pixel classification model")
                return
            else:
                try:
                    return getattr(self, "imagery_model").average_precision_score()
                except Exception as E:
                    print("Load the model first using load()")
        else:
            print("Train the model first using fit()")
            return


all_val_losses, all_train_losses, dice, temp_log_msg = [], [], [], []
BestPerformingModel, best_backbone, best_model = None, None, None

# Build neural network model


class AutoDL:
    """
    Automates the process of model selection, training and hyperparameter tuning of
    arcgis.learn supported deep learning models within a specified time limit.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Required ImageryDataObject. Returned data object from
                            :meth:`~arcgis.learn.prepare_data`  function.
    ---------------------   -------------------------------------------
    total_time_limit        Optional Int. The total time limit in hours for
                            AutoDL training.
                            Default is 2 Hr.
    ---------------------   -------------------------------------------
    mode                    Optional String.
                            Can be "basic" or "advanced".

                            * basic : To be used when the user wants to train all selected networks.

                            * advanced : To be used when the user wants to tune hyper parameters of two
                            best performing models from basic mode.
    ---------------------   -------------------------------------------
    network                 Optional List of str.
                            The list of models that will be used in the training.
                            For eg:
                            Supported Object Detection models:
                            ["SingleShotDetector", "RetinaNet", "FasterRCNN", "YOLOv3", "MaskRCNN", "DETReg" ,"ATSS",
                            "CARAFE", "CascadeRCNN", "CascadeRPN", "DCN", 'Detectors',
                            'DoubleHeads', 'DynamicRCNN', 'EmpiricalAttention', 'FCOS', 'FoveaBox',
                            'FSAF', 'GHM', 'LibraRCNN', 'PaFPN', 'PISA', 'RegNet','RepPoints',
                            'Res2Net', 'SABL', 'VFNet']
                            Supported Pixel Classification models:
                            ["DeepLab", "UnetClassifier", "PSPNetClassifier",
                                "ANN", "APCNet", "CCNet", "CGNet", "HRNet", 'DeepLabV3Plus',
                                'DMNet', 'DNLNet', 'FastSCNN', 'FCN', 'GCNet', 'MobileNetV2',
                                'NonLocalNet','OCRNet', 'PSANet', 'SemFPN', 'UperNet']

    ---------------------   -------------------------------------------
    verbose                 Optional Boolean.
                            To be used to display logs while training the models.
                            Default is True.

    =====================   ===========================================

    :return:
        :class:`~arcgis.learn.AutoDL` Object
    """

    def __init__(
        self,
        data=None,
        total_time_limit=2,
        mode="basic",
        network=None,
        verbose=True,
        **kwargs
    ):
        if "save_evaluated_models" in kwargs:
            self._save_evaluated_models = kwargs["save_evaluated_models"]
        else:
            self._save_evaluated_models = True

        if "output_folder" in kwargs:
            self._output_path = kwargs["output_folder"]
            self._save_to_folder = True
        else:
            self._output_path = data.path
            self._save_to_folder = False

        if verbose:
            self._logger_dict = []

        prepare_data_args = data.arcgis_init_kwargs
        prepare_data_args["batch_size"] = None
        self._data = prepare_data(**prepare_data_args)

        self.verbose = verbose
        algorithms = network
        self._total_training_time = 0
        self._max_image_set = 500
        self._max_epochs = 20
        self._remaining_time = 0
        self._exhaustive_mode_studies = []
        self._epoch_obj = {}
        self._all_algorithms = [
            "DeepLab",
            "UnetClassifier",
            "PSPNetClassifier",
            "ANN",
            "APCNet",
            "CCNet",
            "CGNet",
            "HRNet",
            "DeepLabV3Plus",
            "DMNet",
            "DNLNet",
            "EMANet",
            "FastSCNN",
            "FCN",
            "GCNet",
            "MobileNetV2",
            "NonLocalNet",
            "OCRNet",
            "PSANet",
            "SemFPN",
            "UperNet",
            "SingleShotDetector",
            "MaskRCNN",
            "RetinaNet",
            "FasterRCNN",
            "YOLOv3",
            "DETReg",
            "ATSS",
            "CARAFE",
            "CascadeRPN",
            "CascadeRCNN",
            "DCN",
            "Detectors",
            "DoubleHeads",
            "DynamicRCNN",
            "EmpiricalAttention",
            "FCOS",
            "FoveaBox",
            "FSAF",
            "GHM",
            "LibraRCNN",
            "PaFPN",
            "PISA",
            "RegNet",
            "RepPoints",
            "Res2Net",
            "SABL",
            "VFNet",
            "MaskRCNN",
        ]
        self._all_mm_algorithms = [
            "ANN",
            "APCNet",
            "CCNet",
            "CGNet",
            "HRNet",
            "ATSS",
            "CARAFE",
            "CascadeRCNN",
            "CascadeRPN",
            "DCN",
            "Detectors",
            "DoubleHeads",
            "DynamicRCNN",
            "EmpiricalAttention",
            "FCOS",
            "FoveaBox",
            "FSAF",
            "GHM",
            "LibraRCNN",
            "PaFPN",
            "PISA",
            "RegNet",
            "RepPoints",
            "Res2Net",
            "SABL",
            "VFNet",
            "DeepLabV3Plus",
            "DMNet",
            "DNLNet",
            "EMANet",
            "FastSCNN",
            "FCN",
            "GCNet",
            "MobileNetV2",
            "NonLocalNet",
            "OCRNet",
            "PSANet",
            "SemFPN",
            "UperNet",
            "MaskRCNN",
        ]
        self._train_df = None
        self._average_precision_score_df = None
        self._mIOU_df = None
        self.best_model = None
        self._best_backbone = None
        self._all_losses = {}
        self._max_accuracy = 0
        self._train_callback = train_callback
        self._remaining_time = 0
        self._all_detection_data = [
            "PASCAL_VOC_rectangles",
            "KITTI_rectangles",
            "RCNN_Masks",
        ]
        if total_time_limit < 0.25:
            raise Exception(
                "Total time limit should be greater than or equal to 0.25 hr"
            )
        total_time_limit = total_time_limit * 60
        self._training_mode = mode.lower()
        if self._training_mode == "perform":
            self._training_mode = "advanced"

        if self._training_mode == "basic":
            self._time_in_sec = total_time_limit * 60
        elif self._training_mode == "advanced":
            self._time_in_sec = (total_time_limit * 60) // 2
        else:
            print("Please select a vaild mode for training..")
            return
        self._algos = []
        self._model_type = self._data._dataset_type

        if self._model_type == "Classified_Tiles":
            if algorithms is None:
                algorithms = self.supported_classification_models()
            all_algos = self.supported_classification_models()
            for algo in algorithms:
                if algo not in all_algos:
                    error = algo + " is not a supported classification model."
                    raise Exception(error)
                if algo == "MMSegmentation":
                    self._algos.extend(MMSegmentation.supported_models)
                else:
                    self._algos.append(algo)
        elif self._model_type in self._all_detection_data:
            if algorithms is None:
                if self._model_type != "RCNN_Masks":
                    algorithms = self.supported_detection_models()
                    algorithms.remove("MaskRCNN")
                else:
                    algorithms = ["MaskRCNN"]

            if self._model_type == "RCNN_Masks":
                self._algos.append("MaskRCNN")
                for algo in algorithms:
                    if algo != "MaskRCNN":
                        error = algo + " is not supported for RCNN_Masks dataset type."
                        raise Exception(error)
            else:
                all_algos = self.supported_detection_models()
                for algo in algorithms:
                    if algo not in all_algos:
                        error = algo + " is not a supported Object Detection model."
                        raise Exception(error)
                    if algo == "MMDetection":
                        self._algos.extend(MMDetection.supported_models)
                    if algo == "MaskRCNN":
                        error = (
                            algo + " is only compatible with RCNN_Masks dataset type."
                        )
                        raise Exception(error)
                    else:
                        self._algos.append(algo)
        else:
            raise Exception("Data must be in ESRI defined format")
        model_stats = self._model_stats()

        if "MaskRCNN" in algorithms and mode == "basic":
            raise Exception("MaskRCNN is only supported with advanced mode.")

        finetune_supported = [
            "SingleShotDetector",
            "MaskRCNN",
            "RetinaNet",
            "DETReg",
            "FasterRCNN",
            "PSPNetClassifier",
            "UnetClassifier",
            "DeepLab",
        ]
        all_mm_models = True
        for algo in self._algos:
            if algo in self._all_algorithms:
                self._total_training_time += int(model_stats[algo]["time"])
            if algo in finetune_supported:
                all_mm_models = False

        if all_mm_models and self._training_mode == "advanced":
            self._time_in_sec = self._time_in_sec * 2

        self._total_training_time //= 60
        if self._total_training_time == 0:
            self._total_training_time = 1

        # self._algos = self._sort_algos(self._algos)

        if total_time_limit is None:
            total_time_limit = self._total_training_time

        number_of_images = len(self._data.train_ds) + len(self._data.valid_ds)
        # print(self._total_training_time, number_of_images, self._max_image_set)
        required_time = (
            self._total_training_time * number_of_images
        ) // self._max_image_set

        if required_time == 0:
            required_time = 1

        self._tiles_required = (
            self._max_image_set * total_time_limit
        ) // self._total_training_time

        if self._tiles_required >= number_of_images:
            self._tiles_required = number_of_images

        ## Edit max time here
        if total_time_limit > required_time:
            self._max_epochs = self._get_max_epochs(total_time_limit, required_time)

        ## Max time edit ends
        # if self._tiles_required <= self.batch_size:
        #     self.batch_size = int(self._tiles_required // 2)
        if round(total_time_limit / 60, 2) == 1:
            unit = "hour"
        else:
            unit = "hours"
        print(
            "Given time to process the dataset is:",
            round(total_time_limit / 60, 2),
            unit,
        )
        print(
            "Number of images that can be processed in the given time:",
            self._tiles_required,
        )
        print(
            "Time required to process the entire dataset of",
            len(self._data.train_ds) + len(self._data.valid_ds),
            "images is",
            round((required_time / 60), 2),
            "hours",
        )

    def _get_max_epochs(self, total_time, required_time):
        self._remaining_time = total_time - required_time
        model_stats = self._model_stats()
        model_epochs = 0
        for algo in self._algos:
            mt = model_stats[algo]["time"]
            model_time_required = (
                (mt // 60) * self._tiles_required
            ) // self._max_image_set
            if model_time_required == 0:
                model_time_required = 1
            time_ratio = (model_time_required / required_time) * 100
            model_remaining_time = (time_ratio / 100) * self._remaining_time
            model_epochs = int((20 * model_remaining_time) // model_time_required)
            break

        self._max_epochs += model_epochs
        return self._max_epochs

    def _train_model(
        self,
        model,
        backbone=None,
        epochs=20,
        model_type="classification",
        model_time=1600,
    ):
        """
        Train the AutoDL models.
        """
        self._is_best = False
        start_time = time.time()
        if model_type == "classification":
            mm_model = "MMSegmentation"
        if model_type == "detection":
            mm_model = "MMDetection"
        if self.verbose:
            log_msg = "{date}: Initializing the {network} network.".format(
                date=dt.now().strftime("%d-%m-%Y %H:%M:%S"), network=model
            )
            print(log_msg)
            self._logger_dict.append(log_msg)

        if backbone is None:
            if not self._model_stats()[model]["is_mm"]:
                setattr(self, model, getattr(ag.learn, model)(self._data))
                try:
                    estimated_batch_size = estimate_batch_size(getattr(self, model))
                except:
                    estimated_batch_size = (2, 2)
                callbacks = [
                    self._train_callback(
                        getattr(self, model).learn,
                        self._tiles_required // self._data.batch_size,
                    )
                ]
            else:
                model_with_underscore = [
                    "CascadeRCNN",
                    "CascadeRPN",
                    "EmpiricalAttention",
                    "MobileNetV2",
                    "NonLocalNet",
                    "SemFPN",
                    "DoubleHeads",
                    "DynamicRCNN",
                    "LibraRCNN",
                    "NasFcos",
                ]
                paired_model = [
                    "Cascade_RCNN",
                    "Cascade_RPN",
                    "Empirical_Attention",
                    "MobileNet_V2",
                    "NonLocal_Net",
                    "Sem_FPN",
                    "Double_Heads",
                    "Dynamic_RCNN",
                    "Libra_RCNN",
                    "Nas_Fcos",
                ]

                if model in model_with_underscore:
                    index = model_with_underscore.index(model)
                    setattr(
                        self,
                        model,
                        getattr(ag.learn, mm_model)(
                            self._data, model=paired_model[index]
                        ),
                    )
                else:
                    setattr(
                        self,
                        model,
                        getattr(ag.learn, mm_model)(self._data, model=model.lower()),
                    )
                try:
                    estimated_batch_size = estimate_batch_size(getattr(self, model))
                except:
                    estimated_batch_size = (2, 2)
                callbacks = [
                    self._train_callback(
                        getattr(self, model).learn,
                        self._tiles_required // self._data.batch_size,
                    )
                ]
            backbone = getattr(self, model)._backbone.__name__
        else:
            if not self._model_stats()[model]["is_mm"]:
                setattr(
                    self, model, getattr(ag.learn, model)(self._data, backbone=backbone)
                )
                try:
                    estimated_batch_size = estimate_batch_size(getattr(self, model))
                except:
                    estimated_batch_size = (2, 2)
                callbacks = [
                    self._train_callback(
                        getattr(self, model).learn,
                        self._tiles_required // self._data.batch_size,
                    )
                ]
            else:
                model_with_underscore = [
                    "CascadeRCNN",
                    "CascadeRPN",
                    "EmpiricalAttention",
                    "MobileNetV2",
                    "NonLocalNet",
                    "SemFPN",
                    "DoubleHeads",
                    "DynamicRCNN",
                    "LibraRCNN",
                    "NasFcos",
                ]
                paired_model = [
                    "Cascade_RCNN",
                    "Cascade_RPN",
                    "Empirical_Attention",
                    "MobileNet_V2",
                    "NonLocal_Net",
                    "Sem_FPN",
                    "Double_Heads",
                    "Dynamic_RCNN",
                    "Libra_RCNN",
                    "Nas_Fcos",
                ]

                if model in model_with_underscore:
                    index = model_with_underscore.index(model)
                    setattr(
                        self,
                        model,
                        getattr(ag.learn, mm_model)(
                            self._data, model=paired_model[index]
                        ),
                    )
                else:
                    setattr(
                        self,
                        model,
                        getattr(ag.learn, mm_model)(self._data, model=model.lower()),
                    )
                try:
                    estimated_batch_size = estimate_batch_size(getattr(self, model))
                except:
                    estimated_batch_size = (2, 2)
                callbacks = [
                    self._train_callback(
                        getattr(self, model).learn,
                        self._tiles_required // self._data.batch_size,
                    )
                ]

        if self.verbose:
            log_msg = "{date}: {network} initialized with {bk} backbone".format(
                date=dt.now().strftime("%d-%m-%Y %H:%M:%S"), network=model, bk=backbone
            )
            print(log_msg)
            self._logger_dict.append(log_msg)

        if self.verbose:
            log_msg = "{date}: finding desired batch size for the data object.".format(
                date=dt.now().strftime("%d-%m-%Y %H:%M:%S"),
            )
            print(log_msg)
            self._logger_dict.append(log_msg)
        if estimated_batch_size[0] > 2:
            getattr(self, model)._data.batch_size = estimated_batch_size[0] // 2
        else:
            getattr(self, model)._data.batch_size = 2
        if estimated_batch_size[0] > self._tiles_required:
            getattr(self, model)._data.batch_size = (
                2 ** (int(self._tiles_required) - 1).bit_length() // 2
            )

        if self.verbose:
            log_msg = "{date}: Optimized batch size for {network} with the selected backbone is {lr}".format(
                date=dt.now().strftime("%d-%m-%Y %H:%M:%S"),
                network=model,
                lr=estimated_batch_size[0],
            )
            print(log_msg)
            self._logger_dict.append(log_msg)

        lr_val = getattr(self, model).lr_find(allow_plot=False)
        if self.verbose:
            log_msg = "{date}: Best learning rate for {network} with the selected data is {lr}".format(
                date=dt.now().strftime("%d-%m-%Y %H:%M:%S"), network=model, lr=lr_val
            )
            print(log_msg)
            self._logger_dict.append(log_msg)

            log_msg = "{date}: Fitting {network}".format(
                date=dt.now().strftime("%d-%m-%Y %H:%M:%S"), network=model
            )
            print(log_msg)
            self._logger_dict.append(log_msg)

        try:
            _training_time_ = 0
            while model_time > _training_time_:
                # clear_output(wait=True)
                _start_time = time.time()
                with RedirectedStdout() as out:
                    getattr(self, model).fit(
                        int(epochs),
                        lr=lr_val,
                        early_stopping=True,
                        checkpoint=False,
                        callbacks=callbacks,
                    )
                    init_log = str(out)
                if "early stopping" in init_log:
                    if self.verbose:
                        log_msg = "{date}: Early stopping the {model} ".format(
                            date=dt.now().strftime("%d-%m-%Y %H:%M:%S"),
                            model=model,
                        )
                        print(log_msg)
                        self._logger_dict.append(log_msg)
                    break
                _end_time = time.time()
                _training_time_ = _end_time - _start_time
                model_time -= _training_time_
                epochs = int((model_time * epochs) // _training_time_)
                if model_time > _training_time_:
                    if epochs > 200:
                        epochs == 200
                    if self.verbose:
                        log_msg = "{date}: Time left for {epochs} more epochs, training the {model} again. ".format(
                            date=dt.now().strftime("%d-%m-%Y %H:%M:%S"),
                            epochs=epochs,
                            model=model,
                        )
                        print(log_msg)

                else:
                    break

            if self.verbose:
                clear_output(wait=True)
                all_logs = "\n".join(self._logger_dict)
                print(all_logs)
        except ToolIsCancelled as e:
            raise Exception("Tool is cancelled")
            exit()
        except Exception as e:
            print("Error: ", e)
            end_time = time.time()
            tot_sec = int(end_time - start_time)
            delattr(self, model)
            gc.collect()
            torch.cuda.empty_cache()

            return tot_sec

        if self.verbose:
            log_msg = "{date}: Training completed".format(
                date=dt.now().strftime("%d-%m-%Y %H:%M:%S"), network=model
            )
            print(log_msg)
            self._logger_dict.append(log_msg)
        end_time = time.time()
        # print(metrics)
        if self.verbose:
            log_msg = "{date}: Computing the network metrices".format(
                date=dt.now().strftime("%d-%m-%Y %H:%M:%S")
            )
            print(log_msg)
            self._logger_dict.append(log_msg)
        metrics = getattr(self, model).learn.recorder.get_state()
        self._all_losses[model] = metrics["losses"]
        tot_sec = int(end_time - start_time)
        # self.m = metrics
        #
        t = str(datetime.timedelta(seconds=tot_sec))
        gc.collect()
        torch.cuda.empty_cache()
        train_loss = np.array(metrics["losses"])[-1]
        valid_loss = np.array(metrics["val_losses"])[-1]
        name_time = time.strftime("%Y-%m-%d_%H-%M-%S")
        if model_type == "classification":
            accuracy = np.array(metrics["metrics"])[-1][0]
            miou = getattr(self, model).mIOU()
            miou["Model"] = str(model)
            self._mIOU_df = pd.concat(
                [
                    self._mIOU_df,
                    pd.DataFrame({key: [val] for key, val in miou.items()}),
                ]
            )

            if accuracy >= self._max_accuracy:
                self._is_best = True
                self._max_accuracy = accuracy
            dice = np.array(metrics["metrics"])[-1][1]
            df = pd.DataFrame(
                {
                    "Model": [model],
                    "train_loss": [train_loss],
                    "valid_loss": [valid_loss],
                    "accuracy": [accuracy],
                    "dice": [dice],
                    "lr": [lr_val],
                    "training time": [t],
                    "backbone": [backbone],
                    "timing": [name_time],
                    "optuna_study": [False],
                }
            )
            self._train_df = pd.concat([self._train_df, df], ignore_index=True)
        else:
            avg_precision = getattr(self, model).average_precision_score()
            avg = sum(avg_precision.values()) / len(avg_precision.values())
            avg_precision["Model"] = str(model)
            self._average_precision_score_df = pd.concat(
                [
                    self._average_precision_score_df,
                    pd.DataFrame({key: [val] for key, val in avg_precision.items()}),
                ]
            )
            if avg >= self._max_accuracy:
                self._is_best = True
                self._max_accuracy = avg
            df = pd.DataFrame(
                {
                    "Model": [model],
                    "train_loss": [train_loss],
                    "valid_loss": [valid_loss],
                    "average_precision_score": [avg],
                    "lr": [lr_val],
                    "training time": [t],
                    "backbone": [backbone],
                    "timing": [name_time],
                    "optuna_study": [False],
                }
            )
            self._train_df = pd.concat([self._train_df, df], ignore_index=True)
        if self.verbose:
            log_msg = "{date}: Finished training {network}.".format(
                date=dt.now().strftime("%d-%m-%Y %H:%M:%S"), network=model
            )
            print(log_msg)
            self._logger_dict.append(log_msg)
            log_msg = "{date}: Exiting.".format(
                date=dt.now().strftime("%d-%m-%Y %H:%M:%S")
            )
            print(log_msg)
            self._logger_dict.append(log_msg)

        ## Save the model
        if self.verbose:
            log_msg = "{date}: Saving the model".format(
                date=dt.now().strftime("%d-%m-%Y %H:%M:%S")
            )
            print(log_msg)
            self._logger_dict.append(log_msg)

        # print(self._save_evaluated_models)
        if self._save_evaluated_models:
            if self._save_to_folder:
                getattr(self, model).save(
                    self._output_path
                    + os.sep
                    + "models"
                    + os.sep
                    + "AutoDL_"
                    + str(model)
                    + "_"
                    + backbone
                    + "_"
                    + name_time
                )
                if self.verbose:
                    log_msg = "{date}: model saved at {path}".format(
                        date=dt.now().strftime("%d-%m-%Y %H:%M:%S"),
                        path=os.path.join(
                            self._output_path,
                            "models",
                            "AutoDL_" + str(model) + "_" + backbone + "_" + name_time,
                        ),
                    )
                    print(log_msg)
                    self._logger_dict.append(log_msg)
            else:
                getattr(self, model).save(
                    "AutoDL_" + str(model) + "_" + backbone + "_" + name_time
                )
                if self.verbose:
                    log_msg = "{date}: model saved at {path}".format(
                        date=dt.now().strftime("%d-%m-%Y %H:%M:%S"),
                        path=os.path.join(
                            self._data.path,
                            "models",
                            "AutoDL_" + str(model) + "_" + backbone + "_" + name_time,
                        ),
                    )
                    print(log_msg)
                    self._logger_dict.append(log_msg)

        if self._is_best:
            self.best_model = model
            self._best_backbone = backbone
            setattr(self, "BestPerformingModel", getattr(self, model))

        if not self._model_stats()[model]["is_mm"]:
            setattr(
                self, model + "_backbones", getattr(self, model).supported_backbones
            )
            delattr(self, model)
            gc.collect()
            torch.cuda.empty_cache()

        else:
            delattr(self, model)
            gc.collect()
            torch.cuda.empty_cache()
            if self.verbose:
                log_msg = "{date}: deleting {network} with {bk}".format(
                    date=dt.now().strftime("%d-%m-%Y %H:%M:%S"),
                    network=model,
                    bk=backbone,
                )
                print(log_msg)
                self._logger_dict.append(log_msg)
        return tot_sec

    def fit(self, **kwargs):
        """
        Train the selected networks for the specified number of epochs and using the
        specified learning rates
        """
        self._logger_dict = []
        if self._model_type == "Classified_Tiles":
            m_type = "classification"
            self._train_df = pd.DataFrame(
                columns=[
                    "Model",
                    "train_loss",
                    "valid_loss",
                    "accuracy",
                    "dice",
                    "lr",
                    "training time",
                    "backbone",
                    "timing",
                    "optuna_study",
                ]
            )
        elif self._model_type in self._all_detection_data:
            m_type = "detection"
            self._train_df = pd.DataFrame(
                columns=[
                    "Model",
                    "train_loss",
                    "valid_loss",
                    "average_precision_score",
                    "lr",
                    "training time",
                    "backbone",
                    "timing",
                    "optuna_study",
                ]
            )

        compare_time = self._time_in_sec
        if self.verbose:
            log_msg = "{date}: Selected networks: {networks}".format(
                date=dt.now().strftime("%d-%m-%Y %H:%M:%S"),
                networks=" ".join(self._algos),
            )
            print(log_msg)
            self._logger_dict.append(log_msg)

        for model in self._algos:
            # self._max_epochs = int(self._epoch_obj[model])
            if self.verbose:
                log_msg = "{date}: Current network - {network}. ".format(
                    date=dt.now().strftime("%d-%m-%Y %H:%M:%S"), network=model
                )
                print(log_msg)
                self._logger_dict.append(log_msg)

            model_time = self._model_stats()[model]["time"]
            model_time = (model_time * self._tiles_required) // self._max_image_set
            mt = str(datetime.timedelta(seconds=model_time))
            if self.verbose:
                log_msg = "{date}: Total time alloted to train the {network} model is {network_time}".format(
                    date=dt.now().strftime("%d-%m-%Y %H:%M:%S"),
                    network=model,
                    network_time=mt,
                )
                print(log_msg)
                self._logger_dict.append(log_msg)
            if model_time > compare_time:
                epochs = int((self._max_epochs * compare_time) // model_time)
                if epochs <= 0:
                    epochs = 0
                if self.verbose:
                    log_msg = """{date}: Insufficient time to train the {network} for 20 epochs. {net_epochs} epochs can only be trained in the remaining time.""".format(
                        date=dt.now().strftime("%d-%m-%Y %H:%M:%S"),
                        network=model,
                        net_epochs=epochs,
                    )
                    print(log_msg)
                    self._logger_dict.append(log_msg)
            else:
                epochs = self._max_epochs
                if self.verbose:
                    log_msg = """{date}: Maximum number of epochs will be {net_epochs} to train {network}""".format(
                        date=dt.now().strftime("%d-%m-%Y %H:%M:%S"),
                        network=model,
                        net_epochs=epochs,
                    )
                    self._logger_dict.append(log_msg)
                    print(log_msg)

            if epochs <= 0:
                if self.verbose:
                    log_msg = """{date}: The time left to train the {network} is not sufficent.""".format(
                        date=dt.now().strftime("%d-%m-%Y %H:%M:%S"), network=model
                    )
                    print(log_msg)
                    self._logger_dict.append(log_msg)
                    log_msg = """{date}: Remaining networks will be skipped due to limited time, Stopping the training process.""".format(
                        date=dt.now().strftime("%d-%m-%Y %H:%M:%S")
                    )
                    print(log_msg)
                    self._logger_dict.append(log_msg)
                break

            # tot_sec = self._train_model(
            #     model, epochs=epochs, model_type=m_type, model_time=model_time
            # )

            self.train_basic_model = self._train_model
            tot_sec = self.train_basic_model(
                model, epochs=epochs, model_type=m_type, model_time=model_time
            )
            self.train_basic_model = None
            del self.train_basic_model
            gc.collect()
            torch.cuda.empty_cache()

            compare_time -= tot_sec
        self._dataset_type = m_type
        if m_type == "classification":
            self._train_df = self._train_df.sort_values(
                "accuracy", ascending=False
            ).reset_index(drop=True)
        if m_type == "detection":
            self._train_df = self._train_df.sort_values(
                "average_precision_score", ascending=False
            ).reset_index(drop=True)

        if self._training_mode == "advanced":
            if self.verbose:
                log_msg = """{date}: Entering into exhaustive mode.""".format(
                    date=dt.now().strftime("%d-%m-%Y %H:%M:%S")
                )
                print(log_msg)
                self._logger_dict.append(log_msg)
            compare_time = self._time_in_sec

            # Edited
            top_models = list(self._train_df.head(2)["Model"])
            all_trained_models = list(self._train_df["Model"])

            # top_models = ["SingleShotDetector", "RetinaNet"]
            # all_trained_models = ["SingleShotDetector", "RetinaNet"]

            if self.verbose:
                log_msg = (
                    """{date}: Top two performing models are - {network}""".format(
                        date=dt.now().strftime("%d-%m-%Y %H:%M:%S"),
                        network=" ".join(top_models),
                    )
                )
                print(log_msg)
                self._logger_dict.append(log_msg)

            # edit start here
            # compare_time = 2000
            time_for_each_model = compare_time // 2
            # best_models = top_models
            self._exhaustive_mode_studies = []
            global all_val_losses, all_train_losses, dice, BestPerformingModel, best_backbone, best_model
            global self_obj
            self_obj = self
            counter = 0
            for model in all_trained_models:
                if counter >= 2:
                    break

                all_val_losses = []
                all_train_losses = []
                dice = []
                if self._model_stats()[model]["is_mm"] or model == "YOLOv3":
                    if self.verbose:
                        log_msg = """{date}: {model} does not have additional parameters to tune, skipping.""".format(
                            date=dt.now().strftime("%d-%m-%Y %H:%M:%S"),
                            model=model,
                        )
                        print(log_msg)
                        self._logger_dict.append(log_msg)
                    continue
                self._execute_optuna_study = _train_exhaust_mode
                (
                    current_study,
                    self,
                    all_val_losses,
                    all_train_losses,
                    dice,
                    timing,
                ) = self._execute_optuna_study(
                    self,
                    time_for_each_model,
                    model,
                    all_val_losses,
                    all_train_losses,
                    dice,
                )
                self._execute_optuna_study = None
                del self._execute_optuna_study
                gc.collect()
                torch.cuda.empty_cache()
                df = current_study.trials_dataframe()

                current_study._timing = timing
                df["valid_loss"] = all_val_losses
                df["train_loss"] = all_train_losses
                df["timing"] = timing
                df["optuna_study"] = [True] * len(all_val_losses)
                if self._dataset_type == "classification":
                    df["dice"] = dice
                sorted_df = df.sort_values(by=["value"], ascending=False)
                # sorted_df = sorted_df.drop_duplicates('params_backbones', keep='first')
                sorted_df["duration"] = sorted_df["duration"].apply(
                    lambda x: str(x).split(" ")[-1].split(".")[0]
                )

                if self._dataset_type == "classification":
                    updated_df = {
                        "Model": list([model] * sorted_df.shape[0]),
                        "train_loss": list(sorted_df["train_loss"]),
                        "valid_loss": list(sorted_df["valid_loss"]),
                        "accuracy": list(sorted_df["value"]),
                        "dice": list(sorted_df["dice"]),
                        "lr": list(sorted_df["params_lr"]),
                        "training time": list(sorted_df["duration"]),
                        "backbone": list(sorted_df["params_backbones"]),
                        "timing": list(sorted_df["timing"]),
                        "optuna_study": list(sorted_df["optuna_study"]),
                    }
                else:
                    updated_df = {
                        "Model": list([model] * sorted_df.shape[0]),
                        "train_loss": list(sorted_df["train_loss"]),
                        "valid_loss": list(sorted_df["valid_loss"]),
                        "average_precision_score": list(sorted_df["value"]),
                        "lr": list(sorted_df["params_lr"]),
                        "training time": list(sorted_df["duration"]),
                        "backbone": list(sorted_df["params_backbones"]),
                        "timing": list(sorted_df["timing"]),
                        "optuna_study": list(sorted_df["optuna_study"]),
                    }
                df_new = pd.DataFrame(updated_df)
                self._train_df = pd.concat([self._train_df, df_new], ignore_index=True)
                self._exhaustive_mode_studies.append(current_study)
                counter += 1

                # if BestPerformingModel is not None:
                #     self.BestPerformingModel = BestPerformingModel
                #     self._best_backbone = best_backbone
                #     self.best_model = best_model
                #     self.name_time = time.strftime("%Y-%m-%d_%H-%M-%S")

        if self.verbose:
            log_msg = """{date}: Collating and evaluating model performances.""".format(
                date=dt.now().strftime("%d-%m-%Y %H:%M:%S")
            )
            print(log_msg)
            self._logger_dict.append(log_msg)

        self.name_time = time.strftime("%Y-%m-%d_%H-%M-%S")
        if not self._save_evaluated_models:
            if self._save_to_folder:
                self.BestPerformingModel.save(
                    self._output_path
                    + os.sep
                    + "models"
                    + os.sep
                    + "AutoDL_"
                    + str(self.best_model)
                    + "_"
                    + self._best_backbone
                    + "_"
                    + self.name_time
                )
                if self.verbose:
                    log_msg = "{date}: Saving best performing model at {path}".format(
                        date=dt.now().strftime("%d-%m-%Y %H:%M:%S"),
                        path=os.path.join(
                            self._output_path,
                            "models",
                            "AutoDL_"
                            + str(self.best_model)
                            + "_"
                            + self._best_backbone
                            + "_"
                            + self.name_time,
                        ),
                    )
                    print(log_msg)
                    self._logger_dict.append(log_msg)
            else:
                self.BestPerformingModel.save(
                    "AutoDL_"
                    + str(self.best_model)
                    + "_"
                    + self._best_backbone
                    + "_"
                    + self.name_time
                )
                if self.verbose:
                    log_msg = "{date}: model saved at {path}".format(
                        date=dt.now().strftime("%d-%m-%Y %H:%M:%S"),
                        path=os.path.join(
                            self._data.path,
                            "models",
                            "AutoDL_"
                            + str(model)
                            + "_"
                            + self._best_backbone
                            + "_"
                            + self.name_time,
                        ),
                    )
                    print(log_msg)
                    self._logger_dict.append(log_msg)

        if m_type == "classification":
            self._train_df = self._train_df.sort_values(
                "accuracy", ascending=False
            ).reset_index(drop=True)
        if m_type == "detection":
            self._train_df = self._train_df.sort_values(
                "average_precision_score", ascending=False
            ).reset_index(drop=True)

        if self.verbose:
            log_msg = """{date}: Exiting.""".format(
                date=dt.now().strftime("%d-%m-%Y %H:%M:%S")
            )
            print(log_msg)
            self._logger_dict.append(log_msg)

        _all_executed_model = list(self._train_df["Model"])

        if self._save_to_folder:
            self.BestPerformingModel.save(self._output_path)

        for ex_models in _all_executed_model:
            try:
                delattr(self, ex_models)
                delattr(self, ex_models + "_backbones")
            except AttributeError:
                pass

    def show_results(self, rows=5, threshold=0.25, **kwargs):
        """
        Shows sample results for the model.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional number of rows. By default, 5 rows
                                are displayed.
        =====================   ===========================================

        :return: dataframe
        """
        print(
            "show_results will only show the output from the best performing model: "
            + self.best_model
        )
        plt.show(
            getattr(self, "BestPerformingModel").show_results(
                rows=rows, thresh=threshold, **kwargs
            )
        )
        pass

    def _display_plot(self, x, y):
        plt.bar(x, y)
        plt.xticks(rotation=60)
        plt.show()
        pass

    def score(self, allow_plot=False):
        """
        returns output from AutoDL's model.score(), "average precision score" in case of detection and accuracy in case of classification.
        """
        if self._train_df is None:
            raise Exception("Train a model using fit() before getting the scores.")
        else:
            if allow_plot:
                if self._model_type == "Classified_Tiles":
                    self._display_plot(
                        self._train_df["Model"], self._train_df["accuracy"]
                    )
                else:
                    self._display_plot(
                        self._train_df["Model"],
                        self._train_df["average_precision_score"],
                    )
            df = self._train_df[list(self._train_df.keys())[:-2]]
            return df

    def report(self, allow_plot=False):
        """
        returns a HTML report of the different models trained by AutoDL along with their performance.
        """
        from IPython.display import HTML
        from ._autodl_utils import generate_output_report

        if self._train_df is None:
            raise Exception("Train a model using fit() before getting the scores.")
        else:
            generate_output_report(
                self._train_df,
                self._output_path,
                self._exhaustive_mode_studies,
                self._training_mode,
                self._save_to_folder,
                self._save_evaluated_models,
            )
            # clear_output(wait=True)
            rel_report_path = os.path.join(self._output_path, "README.html")
            return HTML(rel_report_path)

    def average_precision_score(self):
        """
        Calculates the average of the "average precision score" of all classes for selected networks
        """
        if self._model_type == "Classified_Tiles":
            print("This method is not supported with the selected model type")
        elif self._model_type in self._all_detection_data:
            if self._average_precision_score_df is None:
                print("Please train the networks first using fit()")
                return
            cols = ["Model"] + list(self._average_precision_score_df.keys())[:-1]
            return self._average_precision_score_df[cols]
        else:
            print("Datatype not supported!!")

    def mIOU(self):
        """
        Calculates the mIOU of all classes for selected networks
        """
        if self._model_type == "Classified_Tiles":
            try:
                cols = ["Model"] + list(self._mIOU_df.keys())[:-1]
                return self._mIOU_df[cols]
            except Exception as E:
                print(E)
                print("Please train the networks first using fit()")

        elif self._model_type in self._all_detection_data:
            print("This method is not supported with the selected model type")
        else:
            print("Datatype not supported!!")

    def _model_stats(self):
        """
        Shows the model stats
        """
        details = {
            "DeepLab": {
                "time": 1600,
                "is_mm": False,
                "executed": False,
                "params": {
                    "type_list": {
                        "backbones": [
                            "resnet18",
                            "resnet34",
                            "resnet50",
                            "resnet101",
                            "resnet152",
                            "densenet121",
                            "densenet169",
                            "densenet161",
                            "densenet201",
                            "vgg11",
                            "vgg11_bn",
                            "vgg13",
                            "vgg13_bn",
                            "vgg16",
                            "vgg16_bn",
                            "vgg19",
                            "vgg19_bn",
                        ],
                        "dice_loss_average": ["micro", "macro"],
                        "pointrend": [True, False],
                        "class_balancing": [True, False],
                        "mixup": [True, False],
                        "focal_loss": [True, False],
                        "keep_dilation": [True, False],
                    },
                    "type_float": {"dice_loss_fraction": (0, 1)},
                },
            },
            "UnetClassifier": {
                "time": 1600,
                "is_mm": False,
                "executed": False,
                "params": {
                    "type_list": {
                        "backbones": [
                            "resnet18",
                            "resnet34",
                            "resnet50",
                            "resnet101",
                            "resnet152",
                        ],
                        # "backend": ["pytorch", "tensorflow"],
                        "use_unet": [True, False],
                        "pointrend": [True, False],
                        "class_balancing": [True, False],
                        "mixup": [True, False],
                        "focal_loss": [True, False],
                        "keep_dilation": [True, False],
                    },
                    "type_float": {"dice_loss_fraction": (0, 1)},
                },
            },
            "PSPNetClassifier": {
                "time": 1600,
                "is_mm": False,
                "executed": False,
                "params": {
                    "type_list": {
                        "backbones": [
                            "resnet18",
                            "resnet34",
                            "resnet50",
                            "resnet101",
                            "resnet152",
                            "densenet121",
                            "densenet169",
                            "densenet161",
                            "densenet201",
                            "vgg11",
                            "vgg11_bn",
                            "vgg13",
                            "vgg13_bn",
                            "vgg16",
                            "vgg16_bn",
                            "vgg19",
                            "vgg19_bn",
                        ],
                        "dice_loss_average": ["micro", "macro"],
                        "use_unet": [True, False],
                        "pointrend": [True, False],
                        "class_balancing": [True, False],
                        "mixup": [True, False],
                        "focal_loss": [True, False],
                        "keep_dilation": [True, False],
                    },
                    "type_float": {"dice_loss_fraction": (0, 1)},
                },
            },
            "ANN": {
                "time": 1600,
                "is_mm": True,
            },
            "APCNet": {
                "time": 1600,
                "is_mm": True,
            },
            "CCNet": {
                "time": 3500,
                "is_mm": True,
            },
            "CGNet": {
                "time": 700,
                "is_mm": True,
            },
            "HRNet": {
                "time": 4200,
                "is_mm": True,
            },
            "DeepLabV3Plus": {
                "time": 4200,
                "is_mm": True,
            },
            "DMNet": {
                "time": 4200,
                "is_mm": True,
            },
            "DNLNet": {
                "time": 4200,
                "is_mm": True,
            },
            "EMANet": {
                "time": 4200,
                "is_mm": True,
            },
            "FastSCNN": {
                "time": 4200,
                "is_mm": True,
            },
            "FCN": {
                "time": 4200,
                "is_mm": True,
            },
            "GCNet": {
                "time": 4200,
                "is_mm": True,
            },
            "MobileNetV2": {
                "time": 4200,
                "is_mm": True,
            },
            "NonLocalNet": {
                "time": 4200,
                "is_mm": True,
            },
            "OCRNet": {
                "time": 4200,
                "is_mm": True,
            },
            "PSANet": {
                "time": 4200,
                "is_mm": True,
            },
            "SemFPN": {
                "time": 4200,
                "is_mm": True,
            },
            "UperNet": {
                "time": 4200,
                "is_mm": True,
            },
            "SingleShotDetector": {
                "time": 1600,
                "is_mm": False,
                "executed": False,
                "params": {
                    "type_list": {
                        "backbones": [
                            "resnet18",
                            "resnet34",
                            "resnet50",
                            "resnet101",
                            "resnet152",
                            "densenet121",
                            "densenet169",
                            "densenet161",
                            "densenet201",
                            "vgg11",
                            "vgg11_bn",
                            "vgg13",
                            "vgg13_bn",
                            "vgg16",
                            "vgg16_bn",
                            "vgg19",
                            "vgg19_bn",
                            "mobilenet_v2",
                        ],
                        "bias": [True, False],
                        "mixup": [True, False],
                    },
                    "type_float": {
                        "dropout": (0.2, 0.8),
                        "location_loss_factor": (0, 1),
                    },
                },
            },
            "MaskRCNN": {
                "time": 1600,
                "is_mm": False,
                "executed": False,
                "params": {
                    "type_list": {
                        "backbones": [
                            "resnet18",
                            "resnet34",
                            "resnet50",
                            "resnet101",
                            "resnet152",
                        ],
                        "pointrend": [True, False],
                    },
                    "type_int": {
                        "rpn_pre_nms_top_n_train": (1000, 5000),
                        "rpn_pre_nms_top_n_test": (500, 1000),
                    },
                },
            },
            "RetinaNet": {
                "time": 3000,
                "is_mm": False,
                "executed": False,
                "params": {
                    "type_list": {
                        "backbones": [
                            "resnet18",
                            "resnet34",
                            "resnet50",
                            "resnet101",
                            "resnet152",
                        ]
                    }
                },
            },
            "DETReg": {
                "time": 1600,
                "is_mm": False,
                "executed": False,
                "params": {
                    "type_list": {
                        "backbones": [
                            "resnet18",
                            "resnet34",
                            "resnet50",
                            "resnet101",
                            "resnet152",
                        ]
                    }
                },
            },
            "FasterRCNN": {
                "time": 3000,
                "is_mm": False,
                "executed": False,
                "params": {
                    "type_list": {
                        "backbones": [
                            "resnet18",
                            "resnet34",
                            "resnet50",
                            "resnet101",
                            "resnet152",
                        ]
                    },
                    "type_int": {
                        "rpn_pre_nms_top_n_train": (1000, 5000),
                        "rpn_pre_nms_top_n_test": (500, 1000),
                    },
                },
            },
            "YOLOv3": {"time": 6500, "is_mm": False},
            "ATSS": {
                "time": 1650,
                "is_mm": True,
            },
            "CARAFE": {
                "time": 3500,
                "is_mm": True,
            },
            "CascadeRCNN": {
                "time": 700,
                "is_mm": True,
            },
            "CascadeRPN": {
                "time": 4200,
                "is_mm": True,
            },
            "DCN": {
                "time": 4200,
                "is_mm": True,
            },
            "Detectors": {
                "time": 4200,
                "is_mm": True,
            },
            "DoubleHeads": {
                "time": 4200,
                "is_mm": True,
            },
            "DynamicRCNN": {
                "time": 4200,
                "is_mm": True,
            },
            "EmpiricalAttention": {
                "time": 4200,
                "is_mm": True,
            },
            "FCOS": {
                "time": 4200,
                "is_mm": True,
            },
            "FoveaBox": {
                "time": 4200,
                "is_mm": True,
            },
            "FSAF": {
                "time": 4200,
                "is_mm": True,
            },
            "GHM": {
                "time": 4200,
                "is_mm": True,
            },
            "LibraRCNN": {
                "time": 4200,
                "is_mm": True,
            },
            "PaFPN": {
                "time": 4200,
                "is_mm": True,
            },
            "PISA": {
                "time": 4200,
                "is_mm": True,
            },
            "RegNet": {
                "time": 4200,
                "is_mm": True,
            },
            "RepPoints": {
                "time": 4200,
                "is_mm": True,
            },
            "Res2Net": {
                "time": 4200,
                "is_mm": True,
            },
            "SABL": {
                "time": 4200,
                "is_mm": True,
            },
            "VFNet": {
                "time": 4200,
                "is_mm": True,
            },
        }
        return details

    def _sort_algos(self, algorithms):
        """
        Sorts algorithms in a particular order
        """
        sorted_algos = []
        algos = self._all_algorithms
        for algo in algos:
            if algo in algorithms:
                if algo not in sorted_algos:
                    sorted_algos.append(algo)
        return list(sorted_algos)

    def supported_classification_models(self):
        """
        Supported classification models.
        """
        return [
            "DeepLab",
            "UnetClassifier",
            "PSPNetClassifier",
            "ANN",
            "APCNet",
            "CCNet",
            "CGNet",
            "HRNet",
            "DeepLabV3Plus",
            "DMNet",
            "DNLNet",
            "EMANet",
            "FastSCNN",
            "FCN",
            "GCNet",
            "MobileNetV2",
            "NonLocalNet",
            "OCRNet",
            "PSANet",
            "SemFPN",
            "UperNet",
        ]

    def supported_detection_models(self):
        """
        Supported detection models.
        """
        return [
            "SingleShotDetector",
            "RetinaNet",
            "FasterRCNN",
            "YOLOv3",
            "DETReg",
            "ATSS",
            "CARAFE",
            "CascadeRCNN",
            "CascadeRPN",
            "DCN",
            "Detectors",
            "DoubleHeads",
            "DynamicRCNN",
            "EmpiricalAttention",
            "FCOS",
            "FoveaBox",
            "FSAF",
            "GHM",
            "LibraRCNN",
            "PaFPN",
            "PISA",
            "RegNet",
            "RepPoints",
            "Res2Net",
            "SABL",
            "VFNet",
            "MaskRCNN",
        ]

    def lr_find(self):
        print("lr_find() is not supported in AutoDL")



########################################################################
from typing import Optional
import datetime


class CreditManager(object):
    """
    Manages an ArcGIS Online organization's credits for users and sites

    **Example Usage**

    .. code-block:: python

        from arcgis.gis import GIS
        gis = GIS(profile='agol_account')
        cm = gis.admin.credits
        cm.allocate("user1", 100)



    """

    _gis = None
    _con = None
    _portal = None

    # ----------------------------------------------------------------------
    def __init__(self, gis):
        """Constructor"""
        self._gis = gis
        self._portal = gis._portal
        self._con = self._portal.con

    # ----------------------------------------------------------------------
    @property
    def credits(self):
        """returns the current number of credits on the GIS"""
        try:
            return self._gis.properties.availableCredits
        except:
            return 0

    # ----------------------------------------------------------------------
    @property
    def is_enabled(self):
        """
        :return:
            A boolean that shows whether the organization has credit budgeting enabled.
        """
        if "creditAssignments" in self._gis.properties:
            return self._gis.properties.creditAssignments == "enabled"
        else:
            return False

    # ----------------------------------------------------------------------
    def enable(self):
        """
        enables credit allocation on ArcGIS Online
        """
        return self._gis.update_properties({"creditAssignments": "enabled"})

    # ----------------------------------------------------------------------
    def disable(self):
        """
        disables credit allocation on ArcGIS Online
        """
        return self._gis.update_properties({"creditAssignments": "disabled"})

    # ----------------------------------------------------------------------
    @property
    def default_limit(self):
        """
        Gets/Sets the default credit allocation for ArcGIS Online
        """
        return self._gis.properties.defaultUserCreditAssignment

    # ----------------------------------------------------------------------
    @default_limit.setter
    def default_limit(self, value: float):
        """
        Gets/Sets the default credit allocation for ArcGIS Online
        """
        params = {"defaultUserCreditAssignment": value}
        self._gis.update_properties(params)

    # ----------------------------------------------------------------------
    def allocate(self, username: str, credits: Optional[float] = None):
        """
        Allows organization administrators to allocate credits for
        organizational users in ArcGIS Online

        ===========================     ====================================================================
        **Parameter**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        username                        Required string.The name of the user to assign credits to.
        ---------------------------     --------------------------------------------------------------------
        credits                         Optional float. The number of credits to assign to a user. If None
                                        is provided, it sets user to unlimited credits.
        ===========================     ====================================================================

        :return: Boolean. True if successful else False

        """
        if hasattr(username, "username"):
            username = getattr(username, "username")
        if not credits is None:
            params = {
                "f": "json",
                "userAssignments": [{"username": username, "credits": credits}],
            }
            path = "portals/self/assignUserCredits"
            res = self._con.post(path, params)
            if "success" in res:
                return res["success"]
            return res
        else:
            return self.deallocate(username=username)

    # ----------------------------------------------------------------------
    def deallocate(self, username: str):
        """
        Allows organization administrators to set credit limit to umlimited for
        organizational users in ArcGIS Online

        ===========================     ====================================================================
        **Parameter**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        username                        Required string.The name of the user to set to unlimited credits.
        ===========================     ====================================================================

        :return: Boolean. True if successful else False

        """
        if hasattr(username, "username"):
            username = getattr(username, "username")
        params = {"usernames": [username], "f": "json"}
        path = "portals/self/unassignUserCredits"
        res = self._con.post(path, params)
        if "success" in res:
            return res["success"]
        return res

    # ----------------------------------------------------------------------
    def credit_usage(
        self,
        start_time: Optional[datetime.datetime] = None,
        end_time: Optional[datetime.datetime] = None,
        time_frame: str = "week",
    ):
        """
        returns the total credit consumption for a given time period.

        ===================   ===============================================
        **arguements**        **description**
        -------------------   -----------------------------------------------
        start_time            datetime.datetime object. This is the date to
                              start at.
        -------------------   -----------------------------------------------
        end_time              datetime.datetime object. This is the stop time
                              to look for credit consumption. It needs to be
                              at least 1 day previous than then start_time.
        -------------------   -----------------------------------------------
        time_frame            Optional string. is the timeframe report to create.
                              Allowed values: today, week (default), 14days, 30days,
                              60days, 90days, 6months, year

                              If end_time is specified, this parameter is ignored.
        ===================   ===============================================

        returns: dictionary
        """
        from ..._impl.common._utils import local_time_to_online

        if start_time and end_time:
            if (
                isinstance(start_time, datetime.datetime) == False
                or isinstance(end_time, datetime.datetime) == False
            ):
                raise ValueError("start_time and end_time must be datetime objects")

            # calculate time in days between start and end time to assign correct period
            time_elapsed = start_time - end_time
            time_elapsed = time_elapsed.days
            # Convert to timestamps
            if isinstance(start_time, datetime.datetime):
                start_time = str(int(local_time_to_online(dt=start_time)))
            if isinstance(end_time, datetime.datetime):
                end_time = str(int(local_time_to_online(dt=end_time)))
            if time_elapsed <= 1:
                # one day
                period = "1h"
            elif time_elapsed in range(2, 183):
                # less than 6 months
                period = "1d"
            elif time_elapsed in range(183, 265):
                # between 6 months to 1 year
                period = "1w"
            elif time_elapsed >= 365:
                # 1 year or more
                period = "1m"
        elif end_time is None:
            if start_time is None:
                start_time = datetime.datetime.now()
            if time_frame.lower() == "today":
                end_time = start_time - datetime.timedelta(days=1)
                period = "1h"
            elif time_frame.lower() in ["7days", "week"]:
                end_time = start_time - datetime.timedelta(days=7)
                period = "1d"
            elif time_frame.lower() == "14days":
                end_time = start_time - datetime.timedelta(days=14)
                period = "1d"
            elif time_frame.lower() in ["month", "30days"]:
                end_time = start_time - datetime.timedelta(days=30)
                period = "1d"
            elif time_frame.lower() == "60days":
                end_time = start_time - datetime.timedelta(days=60)
                period = "1d"
            elif time_frame.lower() == "90days":
                end_time = start_time - datetime.timedelta(days=90)
                period = "1d"
            elif time_frame.lower() == "6months":
                end_time = start_time - datetime.timedelta(days=180)
                period = "1w"
            elif time_frame.lower() == "year":
                end_time = start_time - datetime.timedelta(days=365)
                period = "1m"
            else:
                raise ValueError("Indicate a valid end_time or time_frame")
            # Convert to timestamps
            end_time = str(int(local_time_to_online(dt=end_time)))
            start_time = str(int(local_time_to_online(dt=start_time)))
        path = "portals/self/usage"
        params = {
            "f": "json",
            "startTime": end_time,
            "endTime": start_time,
            "period": period,
            "groupby": "stype,etype",
            "vars": "credits,num",
        }
        data = self._con.get(path, params)
        res = {}
        for d in data["data"]:
            if d["stype"] in res:
                res[d["stype"]] += sum([float(a[1]) for a in d["credits"]])
            else:
                res[d["stype"]] = sum([float(a[1]) for a in d["credits"]])
        return res



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# reppoints_moment_x101_fpn_dconv_c3-c5_gn-neck+head_2x_coco.py, box AP = 44.2

norm_cfg = dict(type="GN", num_groups=32, requires_grad=True)
model = dict(
    type="RepPointsDetector",
    pretrained="open-mmlab://resnext101_32x4d",
    backbone=dict(
        type="ResNeXt",
        depth=101,
        groups=32,
        base_width=4,
        num_stages=4,
        out_indices=(0, 1, 2, 3),
        frozen_stages=1,
        norm_cfg=dict(type="BN", requires_grad=True),
        style="pytorch",
        dcn=dict(type="DCN", deform_groups=1, fallback_on_stride=False),
        stage_with_dcn=(False, True, True, True),
    ),
    neck=dict(
        type="FPN",
        in_channels=[256, 512, 1024, 2048],
        out_channels=256,
        start_level=1,
        add_extra_convs="on_input",
        norm_cfg=norm_cfg,
        num_outs=5,
    ),
    bbox_head=dict(
        type="RepPointsHead",
        num_classes=80,
        in_channels=256,
        feat_channels=256,
        point_feat_channels=256,
        stacked_convs=3,
        num_points=9,
        gradient_mul=0.1,
        point_strides=[8, 16, 32, 64, 128],
        point_base_scale=4,
        norm_cfg=norm_cfg,
        loss_cls=dict(
            type="FocalLoss", use_sigmoid=True, gamma=2.0, alpha=0.25, loss_weight=1.0
        ),
        loss_bbox_init=dict(type="SmoothL1Loss", beta=0.11, loss_weight=0.5),
        loss_bbox_refine=dict(type="SmoothL1Loss", beta=0.11, loss_weight=1.0),
        transform_method="moment",
    ),
    # training and testing settings
    train_cfg=dict(
        init=dict(
            assigner=dict(type="PointAssigner", scale=4, pos_num=1),
            allowed_border=-1,
            pos_weight=-1,
            debug=False,
        ),
        refine=dict(
            assigner=dict(
                type="MaxIoUAssigner",
                pos_iou_thr=0.5,
                neg_iou_thr=0.4,
                min_pos_iou=0,
                ignore_iof_thr=-1,
            ),
            allowed_border=-1,
            pos_weight=-1,
            debug=False,
        ),
    ),
    test_cfg=dict(
        nms_pre=1000,
        min_bbox_size=0,
        score_thr=0.05,
        nms=dict(type="nms", iou_threshold=0.5),
        max_per_img=100,
    ),
)

checkpoint = "http://download.openmmlab.com/mmdetection/v2.0/reppoints/reppoints_moment_x101_fpn_dconv_c3-c5_gn-neck%2Bhead_2x_coco/reppoints_moment_x101_fpn_dconv_c3-c5_gn-neck%2Bhead_2x_coco_20200329-f87da1ea.pth"
